"""Agent Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import environments

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Agents(scrapi_base.ScrapiBase):
    """Core Class for CX Agent Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id
            self.client_options = self._set_region(agent_id)

    @scrapi_base.api_call_counter_decorator
    def _list_agents_client_request(self, location_id) -> List[
        types.agent.Agent
    ]:
        """Builds the List Agents Request object.

        Args:
          location_id: The GCP Location ID in the following format:
            `projects/<project_id>/locations/<location>`

        Returns:
          List of types.agent.Agent"""

        request = types.agent.ListAgentsRequest()
        request.parent = location_id

        client_options = self._set_region(location_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_agents(request)

        agents = []
        for page in response.pages:
            for agent in page.agents:
                agents.append(agent)

        return agents

    def list_agents(
        self,
        project_id: str,
        location: str = None) -> List[types.Agent]:
        """Get list of all CX agents in a given GCP Region or Project.

        This method allows you to provide a GCP Project ID to retrieve all of
        the CX agents across ALL available GCP region. If the optional location
        ID is provided, the method will only pull the agents for that region.

        Args:
          project_id: The GCP Project ID. Ex: `my-cool-gcp-project`
          location_id: The GCP Location ID. Ex: `global`, `us-central1`, etc.

        Returns:
          List of Agent objects
        """

        if not location:
            region_list = [
                "global",
                "us-central1",
                "us-east1",
                "us-west1",
                "asia-northeast1",
                "asia-south1",
                "australia-southeast1",
                "northamerica-northeast1",
                "europe-west1",
                "europe-west2",
            ]

            agents = []
            for region in region_list:
                location_path = f"projects/{project_id}/locations/{region}"
                agents += self._list_agents_client_request(location_path)

        else:
            location_path = f"projects/{project_id}/locations/{location}"
            agents = self._list_agents_client_request(location_path)

        return agents

    @scrapi_base.api_call_counter_decorator
    def get_agent(self, agent_id: str) -> types.Agent:
        """Retrieves a single CX Agent resource object.

        Args:
          agent_id: The formatted CX Agent ID

        Returns:
          A single CX Agent resource object
        """

        request = types.agent.GetAgentRequest()
        request.name = agent_id

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_agent(request)

        return response

    def get_agent_by_display_name(
        self,
        project_id: str,
        display_name: str,
        location_id: str = None,
        region: str = None
    ) -> types.Agent:
        """Get CX agent in a given GCP project by its human readable
            display name.

        Args:
          project_id: The GCP Project ID as string
          display_name: human-readable display name of CX agent as string
          location_id: Optional. The GCP Project/Location ID, as string
              and in this format:
              `projects/<GCP PROJECT ID>/locations/<LOCATION ID>`
              Improves execution time and resolves conflicts caused
              when multiple agents on different regions have identical
              display names.
          region: Optional. The agent's region ID as string.
              Improves execution time and resolves conflicts caused
              when multiple agents on different regions have identical
              display names.
              Syntax for region ID can be found here:
              https://cloud.google.com/dialogflow/cx/docs/concept/region#avail

        Returns:
          CX agent resource object. If no agent is found, returns None.
        """

        if location_id:
            agent_list = self._list_agents_client_request(location_id)

        elif region:
            agent_list = self._list_agents_client_request(
                f"projects/{project_id}/locations/{region}"
                )
        else:
            agent_list = self.list_agents(project_id=project_id)

        possible_agent = None
        matched_agent = None

        for agent in agent_list:
            if agent.display_name == display_name and not matched_agent:
                matched_agent = agent
            elif agent.display_name == display_name and matched_agent:
                possible_agent = agent
            elif agent.display_name.lower() == display_name.lower():
                possible_agent = agent

        if possible_agent and not matched_agent:
            logging.warning(
                "display_name is case-sensitive. Did you mean \"%s\"?",
                possible_agent.display_name
            )
        elif possible_agent and matched_agent:
            logging.warning(
                '''Found multiple agents with the display name \"%s\".
                 Include location_id or region parameter to resolve cross-region
                 ambiguity.''',
                possible_agent.display_name
            )
            matched_agent = None

        return matched_agent

    @scrapi_base.api_call_counter_decorator
    def create_agent(
        self,
        project_id: str,
        display_name: str = None,
        gcp_region: str = "global",
        obj: types.Agent = None,
        **kwargs,
    ):
        """Create a Dialogflow CX Agent with given display name.

        By default the CX Agent will be created in the project that the user
        is currently authenticated to.
        If the user provides an existing Agent object, a new CX Agent will be
        created based on this object and any other input/kwargs will be
        discarded.

        Args:
          project_id: GCP project id where the CX agent will be created
          display_name: Human readable display name for the CX agent
          gcp_region: GCP region to create CX agent. Defaults to 'global'
          obj: (Optional) Agent object to create new agent from
            Refer to `builders.agents.AgentBuilder` to build one.

        Returns:
          The newly created CX Agent resource object.
        """

        parent = f"projects/{project_id}/locations/{gcp_region}"

        if obj:
            agent_obj = obj
            agent_obj.name = ""

        else:
            if not display_name:
                raise ValueError(
                    "At least display_name or obj should be specified."
                )
            agent_obj = types.Agent(
                display_name=display_name,
                default_language_code="en",
                time_zone="America/Chicago"
            )

            # set optional args as agent attributes
            for key, value in kwargs.items():
                setattr(agent_obj, key, value)

        client_options = self._set_region(parent)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_agent(parent=parent, agent=agent_obj)

        return response


    @scrapi_base.api_call_counter_decorator
    def validate_agent(
        self,
        agent_id: str,
        language_code: str = "en",
        timeout: float = None) -> Dict:
        """Initiates the Validation of the CX Agent or Flow.

        This function will start the Validation feature for the given Agent
        and then return the results as a Dict.

        Args:
          agent_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>
          timeout: (Optional) The timeout for this request

        Returns:
          Dictionary of Validation results for the entire Agent.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.agent.ValidateAgentRequest()
        request.name = agent_id
        request.language_code = language_code

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.validate_agent(request, timeout=timeout)

        val_dict = self.cx_object_to_dict(response)

        return val_dict


    @scrapi_base.api_call_counter_decorator
    def get_validation_result(
        self,
        agent_id: str = None,
        timeout: float = None) -> Dict:
        """Extract Validation Results from CX Validation feature.

        This function will get the LATEST validation result run for the given
        CX Agent or CX Flow. If there has been no validation run on the Agent
        or Flow, no result will be returned. Use `dfcx.validate` function to
        run Validation on an Agent/Flow.

        Passing in the Agent ID will provide ALL validation results for
        ALL flows.
        Passing in the Flow ID will provide validation results for only
        that Flow ID.

        Args:
          agent_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>
          timeout: (Optional) The timeout for this request

        Returns:
          Dictionary of Validation results for the entire Agent.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.agent.GetAgentValidationResultRequest()
        request.name = agent_id + "/validationResult"

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_agent_validation_result(
            request, timeout=timeout
        )

        val_results_dict = self.cx_object_to_dict(response)

        return val_results_dict


    @scrapi_base.api_call_counter_decorator
    def export_agent(
        self,
        agent_id: str,
        gcs_bucket_uri: str,
        environment_display_name: str = None
    ) -> str:
        """Exports the specified CX agent to Google Cloud Storage bucket.

        Args:
          agent_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>
          gcs_bucket_uri: The Google Cloud Storage bucket/filepath to export the
            agent to in the following format:
              `gs://<bucket-name>/<object-name>`
          environment_display_name: (Optional) CX Agent environment display name
            as string. If not set, DRAFT environment is assumed.

        Returns:
          A Long Running Operation (LRO) ID that can be used to
            check the status of the export using
              dfcx_scrapi.core.operations->get_lro()
        """
        request = types.agent.ExportAgentRequest()
        request.name = agent_id
        request.agent_uri = gcs_bucket_uri

        if environment_display_name:
            self._environments = environments.Environments(creds=self.creds)
            possible_environment = self._environments.get_environments_map(
                agent_id=agent_id, reverse=True
            ).get(environment_display_name)
            if possible_environment:
                request.environment = possible_environment
            else:
                raise ValueError(
                    "Invalid environment_display_name."
                    f" {environment_display_name} does not exist!"
                )

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_agent(request)

        return response.operation.name


    @scrapi_base.api_call_counter_decorator
    def restore_agent(self, agent_id: str, gcs_bucket_uri: str) -> str:
        """Restores a CX agent from a gcs_bucket location.

        Currently there is no way to restore back to default
        settings via the api. The feature request for this is logged.

        Args:
          agent_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>
          gcs_bucket_uri: The Google Cloud Storage bucket/filepath to export the
            agent to in the following format:
              `gs://<bucket-name>/<object-name>`

        Returns:
          A Long Running Operation (LRO) ID that can be used to
            check the status of the import using
              dfcx_scrapi.core.operations->get_lro()
        """

        request = types.RestoreAgentRequest()
        request.name = agent_id
        request.agent_uri = gcs_bucket_uri

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.restore_agent(request)

        return response.operation.name

    @scrapi_base.api_call_counter_decorator
    def update_agent(
        self, agent_id: str, obj: types.Agent = None, **kwargs
    ) -> types.Agent:
        """Updates a single Agent object based on provided kwargs.

        Args:
          agent_id: CX Agent ID string in the following format
              projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>
          obj: (Optional) The CX Agent object in proper format. This can also
              be extracted by using the get_agent() method.
          kwargs: You may find a list of agent attributes here:
              https://cloud.google.com/python/docs/reference/dialogflow-cx/
                  latest/google.cloud.dialogflowcx_v3beta1.types.Agent
        Returns:
          The updated CX Agent resource object.
        """

        if obj:
            agent = obj
            agent.name = agent_id
        else:
            agent = self.get_agent(agent_id)

        # set agent attributes to args
        for key, value in kwargs.items():
            setattr(agent, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_agent(agent=agent, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_agent(self, agent_id: str) -> str:
        """Deletes the specified Dialogflow CX Agent.

        Args:
          agent_id: CX Agent ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>

        Returns:
          String "Agent '(agent_id)' successfully deleted."
        """
        client_options = self._set_region(agent_id)
        client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )
        client.delete_agent(name=agent_id)

        return "Agent '{agent_id}' successfully deleted."



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import logging
from typing import List, Dict, Union, Any

from google.cloud.dialogflowcx_v3beta1.types import ResponseMessage
from google.protobuf import struct_pb2
from dfcx_scrapi.builders.builders_common import BuildersCommon

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class ResponseMessageBuilder(BuildersCommon):
    """Base Class for CX ResponseMessage builder."""

    _proto_type = ResponseMessage
    _proto_type_str = "ResponseMessage"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        if self.proto_obj.text:
            resp_type = "text"
            resp_msg = self.proto_obj.text.text
        elif self.proto_obj.payload:
            resp_type = "payload"
            proto_struct = self.proto_obj.payload
            resp_msg = "\n\t".join([
                f"{k}: {v}" for k, v in proto_struct.items()
            ])
        elif self.proto_obj.conversation_success:
            resp_type = "conversation_success"
            proto_struct = self.proto_obj.conversation_success.metadata
            resp_msg = "\n\t".join([
                f"{k}: {v}" for k, v in proto_struct.items()
            ])
        elif self.proto_obj.output_audio_text:
            if self.proto_obj.output_audio_text.text:
                resp_type = "output_audio_text - text"
                resp_msg = self.proto_obj.output_audio_text.text
            elif self.proto_obj.output_audio_text.ssml:
                resp_type = "output_audio_text - ssml"
                resp_msg = self.proto_obj.output_audio_text.ssml
        elif self.proto_obj.live_agent_handoff:
            resp_type = "live_agent_handoff"
            proto_struct = self.proto_obj.live_agent_handoff.metadata
            resp_msg = "\n\t".join([
                f"{k}: {v}" for k, v in proto_struct.items()
            ])
        elif self.proto_obj.play_audio:
            resp_type = "play_audio"
            resp_msg = self.proto_obj.play_audio.audio_uri
        elif self.proto_obj.telephony_transfer_call:
            resp_type = "telephony_transfer_call"
            resp_msg = self.proto_obj.telephony_transfer_call.phone_number

        return (
            f"Response Type: {resp_type}\nMessage:\n\t{resp_msg}"
        )


    def show_response_message(self):
        """Show the proto_obj information."""
        self._check_proto_obj_attr_exist()
        print(self)


    def create_new_proto_obj(
        self,
        response_type: str,
        message: Union[str, List[str], Dict[str, Any]],
        mode: str = None
    ) -> ResponseMessage:
        """Create a ResponseMessage that can be returned by a
        conversational agent.
        ResponseMessages are also used for output audio synthesis.

        Args:
          response_type (str):
            Type of the response message. It should be one of the following:
            'text', 'payload', 'conversation_success', 'output_audio_text',
            'live_agent_handoff', 'play_audio', 'telephony_transfer_call'
          message (str | List[str] | Dict[str, str]):
            The output message. For each response_type
            it should be formatted like the following:
              text --> str | List[str]
                A single message as a string or
                multiple messages as a list of strings
              payload --> Dict[str, Any]
                Any dictionary which its keys are string.
                Dialogflow doesn't impose any structure on the values.
              conversation_success --> Dict[str, Any]
                Any dictionary which its keys are string.
                Dialogflow doesn't impose any structure on the values.
              output_audio_text --> str
                A text or ssml response as a string.
              live_agent_handoff --> Dict[str, Any]
                Any dictionary which its keys are string.
                Dialogflow doesn't impose any structure on the values.
              play_audio --> str
                URI of the audio clip.
                Dialogflow does not impose any validation on this value.
              telephony_transfer_call --> str
                A phone number in E.164 format as a string.
                `<https://en.wikipedia.org/wiki/E.164>`
          mode (str):
            This argument is only applicable for `output_audio_text`.
            It should be one of the following: 'text', 'ssml'

        Returns:
          A ResponseMessage object
        """
        if response_type == "text":
            if isinstance(message, str):
                message = [message]
            elif isinstance(message, list):
                if not all((isinstance(msg, str) for msg in message)):
                    raise ValueError(
                        "Only strings are allowed in message list for `text`."
                    )
            else:
                raise ValueError(
                    "For 'text' message should be"
                    " either a string or a list of strings."
                )

            response_message = ResponseMessage(
                text=ResponseMessage.Text(text=message)
            )
        elif response_type == "payload":
            if not(
                isinstance(message, dict)
                and all((isinstance(key, str) for key in message.keys()))
            ):
                raise ValueError(
                    "For `payload`, message should be"
                    " a dictionary and its keys should be strings."
                )

            proto_struct = struct_pb2.Struct()
            proto_struct.update(message)
            response_message = ResponseMessage(payload=proto_struct)
        elif response_type == "conversation_success":
            proto_struct = struct_pb2.Struct()
            proto_struct.update(message)
            convo_success = ResponseMessage.ConversationSuccess(
                metadata=proto_struct
            )
            response_message = ResponseMessage(
                conversation_success=convo_success
            )
        elif response_type == "output_audio_text":
            if not isinstance(message, str):
                raise ValueError(
                    "For 'output_audio_text', message should be a string."
                )

            if mode == "text":
                output_audio_text = ResponseMessage.OutputAudioText(
                    text=message
                )
            elif mode == "ssml":
                output_audio_text = ResponseMessage.OutputAudioText(
                    ssml=message
                )
            else:
                raise ValueError(
                    "mode should be either 'text' or 'ssml'"
                    " for `output_audio_text`."
                )

            response_message = ResponseMessage(
                output_audio_text=output_audio_text
            )
        elif response_type == "live_agent_handoff":
            if not(
                isinstance(message, dict)
                and all((isinstance(key, str) for key in message.keys()))
            ):
                raise ValueError(
                    "For `live_agent_handoff`, message should be"
                    " a dictionary and its keys should be strings."
                )

            proto_struct = struct_pb2.Struct()
            proto_struct.update(message)
            live_agent_handoff = ResponseMessage.LiveAgentHandoff(
                metadata=proto_struct
            )
            response_message = ResponseMessage(
                live_agent_handoff=live_agent_handoff
            )
        elif response_type == "play_audio":
            if not isinstance(message, str):
                raise ValueError(
                    "For 'play_audio', message should be a valid URI string."
                )

            # TODO: Validate the URI here
            response_message = ResponseMessage(
                play_audio=ResponseMessage.PlayAudio(audio_uri=message)
            )
        elif response_type == "telephony_transfer_call":
            e_164_re_pattern = r"^\+[1-9]\d{1,14}$"
            if not (
                isinstance(message, str)
                and re.search(e_164_re_pattern, message)
            ):
                raise ValueError(
                    "For 'telephony_transfer_call',"
                    " message should be a valid E.164 format phone number."
                )
            transfer_call_obj = ResponseMessage.TelephonyTransferCall(
                phone_number=message
            )
            response_message = ResponseMessage(
                telephony_transfer_call=transfer_call_obj
            )
        else:
            raise ValueError(
                "response_type should be one of the following:"
                "\n['text', 'payload', 'conversation_success',"
                " 'output_audio_text', 'live_agent_handoff', 'play_audio',"
                " 'telephony_transfer_call']"
            )

        self.proto_obj = response_message
        return self.proto_obj



"""Utility file for NLU Utterance Similarity to work with CX resources."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import scann
import gspread
from oauth2client.service_account import ServiceAccountCredentials
from typing import Dict, Set

import tensorflow_hub

from dfcx_scrapi.core import flows
from dfcx_scrapi.core import intents
from dfcx_scrapi.core import pages
from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import transition_route_groups

from google.colab import data_table

data_table.enable_dataframe_formatter()

SHEETS_SCOPE = [
    "https://spreadsheets.google.com/feeds",
    "https://www.googleapis.com/auth/drive",
]


class KonaEmbeddingModel:
    """Download USE4 model and prep for calculating embeddings."""
    def __init__(self):
        module_url = "https://tfhub.dev/google/universal-sentence-encoder/4"
        self.model = tensorflow_hub.load(module_url)

    def embed(self, utterances, batch_size=512):
        """Generates embeddings for a given set of utterances."""
        embeddings = []
        for next_idx in range(0, len(utterances), batch_size):
            batch_utterances = utterances[next_idx: next_idx + batch_size]
            batch_embeddings = self.model(batch_utterances).numpy()
            embeddings.append(batch_embeddings)
        embeddings = np.vstack(embeddings)

        return embeddings

class SheetsLoader:
    """Load data from Google Sheets."""
    def __init__(self, creds_path: str = None):
        sheets_creds = ServiceAccountCredentials.from_json_keyfile_name(
            filename=creds_path,
            scopes=SHEETS_SCOPE,
        )
        self.sheets_client = gspread.authorize(sheets_creds)

    def load_column_from_sheet(self, sheet_name, worksheet_name, column_name):
        """Load a column from a Google Sheets file"""
        try:
            sheet = self.sheets_client.open(sheet_name)
        except gspread.SpreadsheetNotFound as gse:
            raise KeyError(
                f"Couldn't find sheet '{sheet_name}'."
                "Did you share it with your service account?"
            ) from gse

        worksheet = sheet.worksheet(worksheet_name)
        worksheet_data = pd.DataFrame(worksheet.get_all_records())
        try:
            column_data = worksheet_data[column_name]
        except KeyError as err:
            raise KeyError(f"Couldn't find column '{column_name}'") from err

        return column_data.to_numpy()


class NaturalLanguageUnderstandingUtil(scrapi_base.ScrapiBase):
    """Class to generate and analyze embeddings for a page."""

    def __init__(
        self,
        agent_id: str,
        flow_display_name: str,
        page_display_name: str,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds=None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
        )

        print("Loading training data...")
        self._load_data(agent_id, flow_display_name, page_display_name)

        print("Loading embedder...")
        self.embedder = KonaEmbeddingModel()

        print("Generating embeddings for training data...")
        (
            self.training_intents,
            self.training_phrases,
        ) = self._get_training_phrases()
        self.training_embeddings = self.generate_embeddings(
            self.training_phrases
        )

        print("Loading ScaNN searcher...")
        self.searcher = self._build_searcher(self.training_embeddings)

    def _load_data(
        self, agent_id: str, flow_display_name: str, page_display_name: str
    ):
        flow_loader = flows.Flows(creds=self.creds)
        self.flow = flow_loader.get_flow_by_display_name(
            flow_display_name, agent_id
        )

        page_loader = pages.Pages(creds=self.creds)
        page_map = page_loader.get_pages_map(self.flow.name, reverse=True)
        page_id = page_map.get(page_display_name, None)
        if page_id is None:
            raise ValueError(
                f'Page "{page_display_name}" does not exist in the '
                "specified agent."
            )
        self.page = page_loader.get_page(page_id)

        trg_loader = transition_route_groups.TransitionRouteGroups(
            creds=self.creds
        )
        self.trgs = trg_loader.list_transition_route_groups(self.flow.name)
        self.name_to_trg = {i.name: i for i in self.trgs}

        intent_names = self._list_page_intents()
        all_intents = intents.Intents(creds=self.creds).list_intents(agent_id)
        self.intents = list(
            filter(lambda x: x.name in intent_names, all_intents)
        )

    def _list_page_intents(self) -> Set[str]:
        """Lists all intents in-scope at the given page."""
        # See the public doc:
        # https://cloud.google.com/dialogflow/cx/docs/concept/handler#eval
        relevant_intents = set()
        relevant_trgs = set(self.flow.transition_route_groups)
        for tr in self.flow.transition_routes:
            if tr.intent:
                relevant_intents.add(tr.intent)
        relevant_trgs |= set(self.page.transition_route_groups)
        for tr in self.page.transition_routes:
            if tr.intent:
                relevant_intents.add(tr.intent)
        for trg_name in relevant_trgs:
            trg = self.name_to_trg[trg_name]
            for tr in trg.transition_routes:
                if tr.intent:
                    relevant_intents.add(tr.intent)

        return relevant_intents

    def _get_training_phrases(self):
        intent_list = []
        training_phrases = []
        for intent in self.intents:
            for training_phrase in intent.training_phrases:
                phrase_str = "".join([i.text for i in training_phrase.parts])
                training_phrases.append(phrase_str)
                intent_list.append(intent.display_name)

        return np.array(intent_list), np.array(training_phrases)

    def generate_embeddings(self, utterances):
        return self.embedder.embed(utterances)

    def _build_searcher(self, embeddings, num_neighbors=10):
        normalized_dataset = (
            embeddings / np.linalg.norm(embeddings, axis=1)[:, np.newaxis]
        )

        # Use ScaNN brute force. This is fine for up to ~20k points
        searcher = (
            scann.scann_ops_pybind.builder(
                normalized_dataset, num_neighbors, "dot_product"
            )
            .score_brute_force()
            .build()
        )

        return searcher

    def find_similar_phrases(self, utterances):
        embeddings = self.generate_embeddings(utterances)
        nearest_idx, similarities = self.searcher.search_batched(embeddings)

        df = pd.DataFrame(
            {
                "Utterance": utterances,
                "Nearest Training Phrase": self.training_phrases[
                    nearest_idx[:, 0]
                ],
                "Nearest Intent": self.training_intents[nearest_idx[:, 0]],
                "Similarity": similarities[:, 0],
            }
        )

        return df

    def find_new_groups(self, utterances):
        utterances = np.unique(utterances)

        embeddings = self.generate_embeddings(utterances)

        train_nearest_idx, train_similarities = self.searcher.search_batched(
            embeddings
        )

        new_searcher = self._build_searcher(embeddings)
        new_nearest_idx, new_similarities = new_searcher.search_batched(
            embeddings
        )

        # Count how many new utterances are more similar
        # than any training phrase.
        closer_count = np.sum(
            new_similarities > train_similarities[:, :1], axis=1
        )

        # Pull out the largest groups.
        grouped_utterance_ids = set()
        groups = []
        similar_training_phrases = []
        similar_intents = []
        training_phrase_distances = []
        for utterance_idx in np.argsort(closer_count)[::-1]:
            if utterance_idx in grouped_utterance_ids:
                continue

            if closer_count[utterance_idx] < 2:
                break

            group_utterances = []
            for other_idx in new_nearest_idx[
                utterance_idx, : closer_count[utterance_idx]
            ]:
                if other_idx in grouped_utterance_ids:
                    # Some of the utterances in this group were
                    #  assigned to another group already, ignore this group.
                    break
                grouped_utterance_ids.add(other_idx)
                group_utterances.append(utterances[other_idx])
            else:
                # Found a new group, add it.
                match_idx = train_nearest_idx[utterance_idx, 0]
                similar_training_phrases.append(
                    self.training_phrases[match_idx]
                )
                similar_intents.append(self.training_intents[match_idx])
                training_phrase_distances.append(
                    train_similarities[utterance_idx, 0]
                )
                groups.append('"' + ('", "'.join(group_utterances)) + '"')

        df = pd.DataFrame(
            {
                "Utterances": groups,
                "Nearest Training Phrase": similar_training_phrases,
                "Nearest Intent": similar_intents,
                "Similarity": training_phrase_distances,
            }
        )

        return df

    def find_similar_training_phrases_in_different_intents(self):
        num_utterances = len(self.training_phrases)
        all_idx_1 = np.tile(np.arange(num_utterances)[:, None], 10)
        all_idx_2, similarities = self.searcher.search_batched(
            self.training_embeddings
        )

        # Only keep pairs in different intents
        def intents_differ(idx_1, idx_2):
            return self.training_intents[idx_1] != self.training_intents[idx_2]

        different_intent_mask = np.vectorize(intents_differ)(
            all_idx_1, all_idx_2
        )

        mismatch_mask = different_intent_mask & (similarities > 0.8)
        mismatch_idx_1 = all_idx_1[mismatch_mask]
        mismatch_idx_2 = all_idx_2[mismatch_mask]
        mismatch_similarities = similarities[mismatch_mask]

        # Remove any duplicates
        sort_mask = mismatch_idx_1 > mismatch_idx_2
        sort_vals_1 = mismatch_idx_1[sort_mask]
        mismatch_idx_1[sort_mask] = mismatch_idx_2[sort_mask]
        mismatch_idx_2[sort_mask] = sort_vals_1
        (unique_idx_1, unique_idx_2), unique_index = np.unique(
            [mismatch_idx_1, mismatch_idx_2], axis=1, return_index=True
        )
        unique_similarities = mismatch_similarities[unique_index]

        df = (
            pd.DataFrame(
                {
                    "Training phrase 1": self.training_phrases[unique_idx_1],
                    "Training phrase 2": self.training_phrases[unique_idx_2],
                    "Intent 1": self.training_intents[unique_idx_1],
                    "Intent 2": self.training_intents[unique_idx_2],
                    "Similarity": unique_similarities,
                }
            )
            .sort_values("Similarity", ascending=False)
            .reset_index(drop=True)
        )

        return df



"""Grouping of semantically similiar utterances.
For best results replace entities with a parameter name ex,
all countries with the word country or all dates with the word date"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import pandas as pd
from sklearn.cluster import DBSCAN
import tensorflow_hub as hub

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


logging.info("embedder status: downloading")
embed = hub.load("https://tfhub.dev/google/universal-sentence-encoder/4")
logging.info("embedder status: imported")


class SemanticClustering:
    """Grouping semantically similiar utterances for a variety of tasks:
    - Intent identification
    - Reducing bloated intents
    - no-match patterns

    This class takes a set of utterances and groups them
        according to semantic similarity.
    Similarity is determined based on distance between
        instances in a feature array.

    Attributes:
        phrases: Indicates the utterances to be clustered.
            Need to specify format.
    """

    def __init__(self, phrases: pd.DataFrame):
        """Initializes SemanticClustering with a pandas data frame"""
        if "text" not in phrases.columns:
            raise ValueError("Utterances dataframe must have a text column")
        self.phrases = phrases

    @staticmethod
    def _string_cleaner(string):
        """Clean text by removing tokens, punctuation, and applying lower().

        Args:
          string: text string to be cleaned

        Returns:
          Cleaned string
        """
        string = re.sub(r"[^\w\s]", "", string.lower())
        tokens = ["\n", "\r", "\t"]
        for token in tokens:
            string = string.replace(token, "")
        return string

    @staticmethod
    def _single_cluster_algo(
        data,
        eps=0.7,
        min_samples=2,
        metric="cosine",
        metric_params: dict = None,
        algorithm="auto",
        leaf_size=30,
        power=None,
        n_jobs=-1,
    ):
        """Cluster phrases using a model with set hyperparameters

        Args:
          data: DataFrame to cluster
          eps: max distance between two points for them to be
            considered in same neighborhood.
          min_samples: minimum number of samples a cluster can have
          metric: metric for measuring distance between instances
            in a feature array.
          metric_params: additional keywords for metric function
          algorithm: algo used to compute pointwise distances and
            find nearest neighbors.
          leaf-size: only passed to BallTree or cKDTree algorithms.
          p: power of Minkowski metric to calculate distance between points.
            DEFAULT = 2 (Euclidean distance)
          n_jobs: number of parallel jobs to run. -1 means all processors

        Returns:
          Input data with associated clusters by the text column.
        """

        input_data = list(data["cleaned_text"])
        vectors = embed(input_data)
        model = DBSCAN(
            eps=eps,
            min_samples=min_samples,
            metric=metric,
            metric_params=metric_params,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=power,
            n_jobs=n_jobs,
        ).fit(vectors)
        data["cluster"] = model.labels_
        return data

    def _run_data_pipeline(self):
        clean_data = self.phrases.copy()
        clean_data["text"] = clean_data["text"].astype(str)
        clean_data["cleaned_text"] = clean_data.apply(
            lambda x: self._string_cleaner(x["text"]),
            axis=1,
        )
        self.clean_data = clean_data

    def cluster(
        self,
        stop_threshold: float = 0.5,
        max_rounds=50,
        iterator=0.05,
        min_samples=2,
        start_eps=0.1,
        metric="cosine",
        metric_params: dict = None,
        algorithm="auto",
        leaf_size=30,
        power=None,
        n_jobs=-1,
    ):
        """Cluster phrases using a model with set hyperparameters
            for the entire dataset.

        User can set stop metrics and multiple models will be generated
            with increasing neighborhood sizes.

        Args:
          stop_threshold: Percentage of data which can be in no cluster
            to signify that new models can stop being created.
          max_rounds: maximum number of rounds that take place of
            trying new model hyperparameters to get to the stop_threshold.
          iterator: eps value to change by in each round.
          start_eps: eps value to run on the first algo.
          min_samples: minimum number of samples a cluster can have
          metric: metric for measuring distance between instances
            in a feature array.
          metric_params: additional keywords for metric function
          algorithm: algo used to compute pointwise distances and find
            nearest neighbors.
          leaf-size: only passed to BallTree or cKDTree algorithms.
          p: power of Minkowski metric to calculate distance between points.
            DEFAULT = 2 (Euclidean distance)
          n_jobs: number of parallel jobs to run. -1 means all processors

        Returns:
          clustered: DataFrame of clustered data.
        """

        if not hasattr(self, "transformed_data"):
            self._run_data_pipeline()

        unclustered = self.clean_data
        instances, unclustered_count = (
            len(unclustered),
            len(unclustered),
        )
        clustered = pd.DataFrame()
        eps, max_cluster, cluster_round = (
            start_eps,
            0,
            1,
        )

        while (
            float(unclustered_count) / float(instances)
        ) > stop_threshold and cluster_round < max_rounds:
            cluster_attempt = self._single_cluster_algo(
                unclustered,
                eps=eps,
                min_samples=min_samples,
                metric=metric,
                metric_params=metric_params,
                algorithm=algorithm,
                leaf_size=leaf_size,
                power=power,
                n_jobs=n_jobs,
            )

            clustered_this_round = pd.DataFrame()
            clustered_this_round = cluster_attempt.copy()[
                cluster_attempt["cluster"] != -1
            ]

            if clustered_this_round.empty is False:
                clusters_refactored = clustered_this_round.apply(
                    lambda x: x["cluster"] + max_cluster,
                    axis=1,
                )
                clustered_this_round = clustered_this_round.drop(
                    columns=["cluster"]
                )
                clustered_this_round.insert(
                    0,
                    "cluster",
                    clusters_refactored,
                )
                clustered_this_round.insert(0, "eps", eps)
                clustered_this_round.insert(0, "round", cluster_round)
                clustered = pd.concat([clustered, clustered_this_round])
                unclustered = cluster_attempt.copy()[
                    cluster_attempt["cluster"] == -1
                ]
                max_cluster = clustered["cluster"].max() + 1
                unclustered_count = len(unclustered)

            eps += iterator
            cluster_round += 1

            unclustered_stats = round(
                        float(unclustered_count) / float(instances) * 100,
                        0,
                    )

            print(
                f"round: {cluster_round} unclusterd: {unclustered_stats}%\t\t"
                ,end="\r",
            )

        if clustered.empty:
            logging.info(
                "no clusters found, try increasing stop_threshold or max_rounds"
            )
            return clustered

        unclustered = unclustered.drop(columns="cluster")
        clustered = pd.concat(
            [clustered.sort_values(by="cluster", ascending=True), unclustered])

        if cluster_round > max_rounds:
            logging.info("max clutering rounds reached before stop threshold")

        return clustered



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass
from typing import List, Union

from google.cloud.dialogflowcx_v3beta1.types import Page
from google.cloud.dialogflowcx_v3beta1.types import Form
from google.cloud.dialogflowcx_v3beta1.types import Fulfillment
from google.cloud.dialogflowcx_v3beta1.types import TransitionRoute
from google.cloud.dialogflowcx_v3beta1.types import EventHandler
from dfcx_scrapi.builders.builders_common import BuildersCommon
from dfcx_scrapi.builders.routes import TransitionRouteBuilder
from dfcx_scrapi.builders.routes import EventHandlerBuilder
from dfcx_scrapi.builders.fulfillments import FulfillmentBuilder

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class PageBuilder(BuildersCommon):
    """Base Class for CX Page builder."""

    _proto_type = Page
    _proto_type_str = "Page"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        return (
            f"Basic Information:\n{'='*25}\n{self._show_basic_info()}"
            f"\n\n\nParameters:\n{'='*25}\n{self._show_parameters()}"
            f"\n\n\nTransitionRoutes:\n{'='*25}"
            f"\n{self._show_transition_routes()}"
            f"\n\n\nEventHandlers:\n{'='*25}\n{self._show_event_handlers()}"
            f"\n\n\nTransitoinRouteGroups:\n{'='*25}"
            f"\n{self._show_transition_route_groups()}")


    def _show_basic_info(self) -> str:
        """String representation for the basic information of proto_obj."""
        self._check_proto_obj_attr_exist()

        entry_fulfillment_str = str(
            FulfillmentBuilder(self.proto_obj.entry_fulfillment)
        )
        return (
            f"display_name: {self.proto_obj.display_name}"
            f"\nentry_fulfillment:\n\n{entry_fulfillment_str}"
        )


    def _show_parameters(self) -> str:
        """String representation for the parameters of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            (
                f"display_name: {param.display_name}"
                f"\n\tentity_type: {param.entity_type}"
                f"\n\trequired: {param.required}"
                f"\n\tis_list: {param.is_list}"
                f"\n\treadct: {param.redact}"
                f"\n\tdefault_value: {param.default_value}"
            )
            for param in self.proto_obj.form.parameters
        ])


    def _show_transition_routes(self) -> str:
        """String representation for the transition routes of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"TransitionRoute {i+1}:\n{str(TransitionRouteBuilder(tr))}"
            f"\n{'*'*20}\n"
            for i, tr in enumerate(self.proto_obj.transition_routes)
        ])


    def _show_event_handlers(self) -> str:
        """String representation for the event handlers of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"EventHandler {i+1}:\n{str(EventHandlerBuilder(eh))}\n{'*'*20}\n"
            for i, eh in enumerate(self.proto_obj.event_handlers)
        ])


    def _show_transition_route_groups(self) -> str:
        """String representation for the transition route groups of proto_obj"""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"TransitionRouteGroup {i+1}: {trg_id}"
            for i, trg_id in enumerate(self.proto_obj.transition_route_groups)
        ])


    def show_page_info(
        self, mode: str = "whole"
    ) -> None:
        """Show the proto_obj information.

        Args:
          mode (str):
            Specifies what part of the page to show.
              Options:
              ['basic', 'whole', 'parameters',
              'routes' or 'transition routes',
              'route groups' or 'transition route groups',
              'events' or 'event handlers'
              ]
        """
        self._check_proto_obj_attr_exist()

        if mode == "basic":
            print(self._show_basic_info())
        elif mode == "parameters":
            print(self._show_parameters())
        elif mode in ["routes", "transition routes"]:
            print(self._show_transition_routes())
        elif mode in ["route groups", "transition route groups"]:
            print(self._show_transition_route_groups())
        elif mode in ["events", "event handlers"]:
            print(self._show_event_handlers())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                "['basic', 'whole', 'parameters',"
                " 'routes', 'transition routes',"
                " 'route groups', 'transition route groups',"
                " 'events', 'event handlers']"
            )


    def show_stats(self) -> None:
        """Provide some stats about the Page."""
        self._check_proto_obj_attr_exist()

        stats_instance = PageStats(self.proto_obj)
        stats_instance.generate_stats()


    def create_new_proto_obj(
        self,
        display_name: str,
        entry_fulfillment: Fulfillment = None,
        overwrite: bool = False
    ) -> Page:
        """Create a new Page.

        Args:
          display_name (str):
            Required. The human-readable name of the
            page, unique within the flow.
          entry_fulfillment (Fulfillment):
            The fulfillment to call when the session is entering the page.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already contains a Page.

        Returns:
          A Page object stored in proto_obj.
        """
        # Types error checking
        if not (display_name and isinstance(display_name, str)):
            raise ValueError("`display_name` should be a nonempty string.")
        if (entry_fulfillment and
            not isinstance(entry_fulfillment, Fulfillment)):
            raise ValueError(
                "The type of `entry_fulfillment` should be a Fulfillment."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains a Page."
                " If you wish to overwrite it, pass `overwrite` as True."
            )
        # Create the Page
        if overwrite or not self.proto_obj:
            if not entry_fulfillment:
                entry_fulfillment = Fulfillment()

            self.proto_obj = Page(
                display_name=display_name,
                entry_fulfillment=entry_fulfillment
            )

        return self.proto_obj


    def add_parameter(
        self,
        display_name: str,
        entity_type: str,
        initial_prompt_fulfillment: Fulfillment,
        required: bool = True,
        default_value: str = None,
        is_list: bool = False,
        redact: bool = False,
        reprompt_event_handlers: Union[EventHandler, List[EventHandler]] = None
    ) -> Page:
        """Add a parameter to collect from the user.

        Args:
          display_name (str):
            Required. The human-readable name of the
            parameter, unique within the form.
          entity_type (str):
            Required. The entity type of the parameter. Format:
            ``projects/-/locations/-/agents/-/
              entityTypes/<System Entity Type ID>``
            for system entity types (for example,
            ``projects/-/locations/-/agents/-/entityTypes/sys.date``),
            or
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/entityTypes/<Entity Type ID>``
            for developer entity types.
          initial_prompt_fulfillment (Fulfillment):
            Required. The fulfillment to provide the initial prompt that
            the agent can present to the user in order to fill the parameter.
          required (bool):
            Indicates whether the parameter is required.
            Optional parameters will not trigger prompts;
            however, they are filled if the user specifies
            them. Required parameters must be filled before
            form filling concludes.
          default_value (str):
            The default value of an optional parameter.
            If the parameter is required, the default value
            will be ignored.
          is_list (bool):
            Indicates whether the parameter represents a
            list of values.
          redact (bool):
            Indicates whether the parameter content should be redacted
            in log. If redaction is enabled, the parameter content will
            be replaced by parameter name during logging.
          reprompt_event_handlers (EventHandler | List[EventHandler]):
            The handlers for parameter-level events, used to provide reprompt
            for the parameter or transition to a different page/flow.
            The supported events are:
              - ``sys.no-match-<N>``, where N can be from 1 to 6
              - ``sys.no-match-default``
              - ``sys.no-input-<N>``, where N can be from 1 to 6
              - ``sys.no-input-default``
              - ``sys.invalid-parameter``

            ``initial_prompt_fulfillment`` provides the first prompt for
            the parameter.

            If the event handler for the corresponding event can't be found on
            the parameter, ``initial_prompt_fulfillment`` will be re-prompted.

        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Types error checking
        if not (display_name and isinstance(display_name, str)):
            raise ValueError("`display_name` should be a nonempty string.")
        if not (entity_type and isinstance(entity_type, str)):
            raise ValueError("`entity_type` should be a valid entity type id.")
        if not (initial_prompt_fulfillment and
            isinstance(initial_prompt_fulfillment, Fulfillment)):
            raise ValueError(
                "`initial_prompt_fulfillment` should be a Fulfillment."
            )
        if not(
            isinstance(required, bool) and
            isinstance(is_list, bool) and
            isinstance(redact, bool)
        ):
            raise ValueError(
                "`is_list`, `required`, and `redact` should be bool."
            )
        if reprompt_event_handlers:
            self._is_type_or_list_of_types(
                reprompt_event_handlers, EventHandler, "reprompt_event_handlers"
            )
            if not isinstance(reprompt_event_handlers, list):
                reprompt_event_handlers = [reprompt_event_handlers]

        if required:
            the_param = Form.Parameter(
                display_name=display_name,
                required=required,
                entity_type=entity_type,
                is_list=is_list,
                redact=redact
            )
        else:
            if not isinstance(default_value, str):
                raise ValueError("`default_value` should be a string.")
            the_param = Form.Parameter(
                display_name=display_name,
                required=required,
                entity_type=entity_type,
                is_list=is_list,
                redact=redact,
                default_value=default_value
            )

        the_param.fill_behavior = Form.Parameter.FillBehavior(
            initial_prompt_fulfillment=initial_prompt_fulfillment,
            reprompt_event_handlers=reprompt_event_handlers
        )
        self.proto_obj.form.parameters.append(the_param)

        return self.proto_obj


    def add_transition_route(
        self,
        transition_routes: Union[TransitionRoute, List[TransitionRoute]]
    ) -> Page:
        """Add single or multiple TransitionRoutes to the Page.

        Args:
          transition_routes (TransitionRoute | List[TransitionRoute]):
            A single or list of TransitionRoutes to add
            to the Page existing in proto_obj.
        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type/Error checking
        self._is_type_or_list_of_types(
            transition_routes, TransitionRoute, "transition_routes"
        )

        if not isinstance(transition_routes, list):
            transition_routes = [transition_routes]
        self.proto_obj.transition_routes.extend(transition_routes)

        return self.proto_obj


    def add_event_handler(
        self,
        event_handlers: Union[EventHandler, List[EventHandler]]
    ) -> Page:
        """Add single or multiple EventHandlers to the Page.

        Args:
          event_handlers (EventHandler | List[EventHandler]):
            A single or list of EventHandler to add
            to the Page existing in proto_obj.
        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type/Error checking
        self._is_type_or_list_of_types(
            event_handlers, EventHandler, "event_handlers"
        )

        if not isinstance(event_handlers, list):
            event_handlers = [event_handlers]
        self.proto_obj.event_handlers.extend(event_handlers)

        return self.proto_obj


    def add_transition_route_group(
        self,
        transition_route_groups: Union[str, List[str]]
    ) -> Page:
        """Add single or multiple TransitionRouteGroups to the Page.

        Args:
          transition_route_groups (str | List[str]):
            A single or list of TransitionRouteGroup's id to add
            to the Page existing in proto_obj. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>/transitionRouteGroups/<TransitionRouteGroup ID>``.
        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type/Error checking
        self._is_type_or_list_of_types(
            transition_route_groups, str, "transition_route_groups"
        )

        if not isinstance(transition_route_groups, list):
            transition_route_groups = [transition_route_groups]
        self.proto_obj.transition_route_groups.extend(transition_route_groups)

        return self.proto_obj


    def remove_parameter(
        self,
        display_name: Union[str, List[str]]
    ) -> Page:
        """Remove single or multiple parameters from the Page.

        Args:
          display_name (str | List[str]):
            A string or a list of strings corresponding to
            the name of the parameter(s).

        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Types error checking
        if not display_name:
            raise ValueError("`display_name` should not be empty.")
        self._is_type_or_list_of_types(display_name, str, "display_name")

        if not isinstance(display_name, list):
            display_name = [display_name]

        new_params = [
            param
            for param in self.proto_obj.form.parameters
            if param.display_name not in display_name
        ]
        self.proto_obj.form.parameters = new_params

        return self.proto_obj


    def remove_transition_route(
        self,
        transition_route: TransitionRoute = None,
        intent: str = None,
        condition: str = None
    ) -> Page:
        """Remove a transition route from the Page.

        At least one of the `transition_route`, `intent`, or `condition` should
        be specfied.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute to remove from the Page.
            intent (str):
              TransitionRoute's intent that should be removed from the Page.
            condition (str):
              TransitionRoute's condition that should be removed from the Page.

        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        new_routes = []
        for tr in self.proto_obj.transition_routes:
            if self._match_transition_route(
                transition_route=tr, target_route=transition_route,
                intent=intent, condition=condition
            ):
                continue
            new_routes.append(tr)
        self.proto_obj.transition_routes = new_routes

        return self.proto_obj


    def remove_event_handler(
        self,
        event_handlers: Union[EventHandler, List[EventHandler]] = None,
        event_names: Union[str, List[str]] = None
    ) -> Page:
        """Remove single or multiple EventHandlers from the Page.

        Args:
          event_handlers (EventHandler | List[EventHandler]):
            A single or list of EventHandler to remove
              from the Page existing in proto_obj.
            Only one of the `event_handlers` and
              `event_names` should be specified.
          event_names (str | List[str]):
            A single or list of EventHandler's event names corresponding to the
              EventHandler to remove from the Page existing in proto_obj.
            Only one of the `event_handlers` and
              `event_names` should be specified.

        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        if event_handlers and event_names:
            raise UserWarning(
                "Only one of the `event_handlers` and "
                "`event_names` should be specified."
            )
        if event_handlers:
            new_ehs = self._find_unmatched_event_handlers(event_handlers)
        elif event_names:
            new_ehs = self._find_unmatched_event_handlers_by_name(event_names)
        else:
            raise UserWarning(
                "At least one of the `event_handlers` and "
                "`event_names` should be specified."
            )

        self.proto_obj.event_handlers = new_ehs

        return self.proto_obj


    def remove_transition_route_group(
        self,
        transition_route_groups: Union[str, List[str]]
    ) -> Page:
        """Remove single or multiple TransitionRouteGroups from the Page.

        Args:
          transition_route_groups (str | List[str]):
            A single or list of TransitionRouteGroup's id to remove
            from the Page existing in proto_obj. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>/transitionRouteGroups/<TransitionRouteGroup ID>``.

        Returns:
          A Page object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        self._is_type_or_list_of_types(
            transition_route_groups, str, "transition_route_groups"
        )

        if not isinstance(transition_route_groups, list):
            transition_route_groups = [transition_route_groups]

        new_trgs = [
            trg
            for trg in self.proto_obj.transition_route_groups
            if trg not in transition_route_groups
        ]
        self.proto_obj.transition_route_groups = new_trgs

        return self.proto_obj



@dataclass
class PageStats():
    """A class for tracking the stats of CX Page object."""
    page_proto_obj: Page

    # Entry Fulfillment
    has_entry_fulfill: bool = False

    # Parameters
    parameters_count: int = 0
    parameters_with_event_handler_count: int = 0
    parameters_with_webhook_fulfill_count: int = 0
    parameters_ratio: int = 0

    # Transition Routes
    transition_routes_count: int = 0
    routes_with_fulfill_count: int = 0
    routes_with_webhook_fulfill_count: int = 0
    intent_routes_count: int = 0
    cond_routes_count: int = 0
    intent_and_cond_routes_count: int = 0

    # Event Handlers
    event_handlers_count: int = 0
    events_with_fulfill_count: int = 0
    events_with_webhook_fulfill_count: int = 0

    # Transition Route Groups
    transition_route_groups_count: int = 0


    def calc_transition_route_stats(self):
        """Calculating TransitionRoute related stats."""
        self.transition_routes_count = len(
            self.page_proto_obj.transition_routes
        )
        for tr in self.page_proto_obj.transition_routes:
            if tr.trigger_fulfillment:
                self.routes_with_fulfill_count += 1
                fb = FulfillmentBuilder(tr.trigger_fulfillment)
                if fb.has_webhook():
                    self.routes_with_webhook_fulfill_count += 1
            if tr.intent and tr.condition:
                self.intent_and_cond_routes_count += 1
            elif tr.intent and not tr.condition:
                self.intent_routes_count += 1
            elif not tr.intent and tr.condition:
                self.cond_routes_count += 1

    def create_transition_route_str(self) -> str:
        """String representation of TransitionRoutes stats."""
        transition_routes_str = (
            f"# of Transition Routes: {self.transition_routes_count}"
        )
        routes_with_fulfill_str = (
            f"# of routes with fulfillment: {self.routes_with_fulfill_count}"
        )
        routes_with_webhook_fulfill_str = (
            "# of routes uses webhook for fulfillment:"
            f" {self.routes_with_webhook_fulfill_count}"
        )
        intent_routes_str = f"# of intent routes: {self.intent_routes_count}"
        cond_routes_str = f"# of condition routes: {self.cond_routes_count}"
        intent_and_cond_routes_str = (
            "# of intent and condition routes:"
            f" {self.intent_and_cond_routes_count}"
        )

        return (
            f"{transition_routes_str}\n\t{intent_routes_str}"
            f"\n\t{cond_routes_str}\n\t{intent_and_cond_routes_str}"
            f"\n\t{routes_with_fulfill_str}"
            f"\n\t{routes_with_webhook_fulfill_str}"
        )


    def calc_event_handler_stats(self):
        """Calculating EventHandler related stats."""
        self.event_handlers_count = len(self.page_proto_obj.event_handlers)
        for eh in self.page_proto_obj.event_handlers:
            fb = FulfillmentBuilder(eh.trigger_fulfillment)
            if fb.has_webhook():
                self.events_with_webhook_fulfill_count += 1
            if eh.trigger_fulfillment:
                self.events_with_fulfill_count += 1

    def create_event_handler_str(self) -> str:
        """String representation of EventHandlers stats."""
        event_handlers_str = f"# of Event Handlers: {self.event_handlers_count}"
        events_with_fulfill_str = (
            "# of Event Handlers with fulfillment:"
            f" {self.events_with_fulfill_count}"
        )
        events_with_webhook_fulfill_str = (
            "# of Event Handlers uses webhook for fulfillment:"
            f" {self.events_with_webhook_fulfill_count}"
        )

        return (
            f"{event_handlers_str}\n\t{events_with_fulfill_str}"
            f"\n\t{events_with_webhook_fulfill_str}"
        )


    def calc_parameter_stats(self):
        """Calculating Parameter related stats."""
        self.parameters_count = len(self.page_proto_obj.form.parameters)
        for param in self.page_proto_obj.form.parameters:
            fb = FulfillmentBuilder(
                param.fill_behavior.initial_prompt_fulfillment
            )
            if fb.has_webhook():
                self.parameters_with_webhook_fulfill_count += 1
            if param.fill_behavior.reprompt_event_handlers:
                self.parameters_with_event_handler_count += 1

        if self.parameters_count != 0:
            self.parameters_ratio = (
                self.parameters_with_event_handler_count/self.parameters_count
            )

    def create_parameter_str(self) -> str:
        """String representation of Page's parameters stats."""
        parameters_str = f"# of Parameters: {self.parameters_count}"
        parameters_with_event_handler_str = (
            "# of Parameters with Event Handlers:"
            f" {self.parameters_with_event_handler_count}"
            f" (Ratio: {self.parameters_ratio})"
        )
        parameters_with_webhook_fulfill_str = (
            "# of Parameters uses webhook for fulfillment:"
            f" {self.parameters_with_webhook_fulfill_count}"
        )

        return (
            f"{parameters_str}\n\t{parameters_with_event_handler_str}"
            f"\n\t{parameters_with_webhook_fulfill_str}"
        )


    def create_entry_fulfillment_str(self) -> str:
        """String representation for Page's Entry Fulfillment."""
        self.has_entry_fulfill = bool(self.page_proto_obj.entry_fulfillment)
        return f"Has entry fulfillment: {self.has_entry_fulfill}"


    def create_transition_route_group_str(self) -> str:
        """String representation of TransitionRouteGroup stats."""
        self.transition_route_groups_count = len(
            self.page_proto_obj.transition_route_groups
        )
        return (
            "# of Transition Route Groups:"
            f" {self.transition_route_groups_count}"
        )


    def generate_stats(self):
        """Generate stats for the Page."""
        self.calc_parameter_stats()
        self.calc_transition_route_stats()
        self.calc_event_handler_stats()

        has_entry_fulfill_str = self.create_entry_fulfillment_str()
        params_stats_str = self.create_parameter_str()
        routes_stats_str = self.create_transition_route_str()
        events_stats_str = self.create_event_handler_str()
        route_groups_stats_str = self.create_transition_route_group_str()

        out = (
            f"{has_entry_fulfill_str}\n{params_stats_str}\n{routes_stats_str}"
            f"\n{events_stats_str}\n{route_groups_stats_str}"
        )
        print(out)



"""Utiliity functions to provide Agent Stats for a Dialogflow CX agent."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.intents import Intents
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.pages import Pages
from dfcx_scrapi.core.entity_types import EntityTypes
from dfcx_scrapi.core.transition_route_groups import TransitionRouteGroups

class StatsUtil(ScrapiBase):
    """A util class to provide common stats for a CX Agent."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id=None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id

        self._agents_tracker = Agents(
            creds=self.creds, agent_id=self.agent_id)
        self._intents_tracker = Intents(
            creds=self.creds, agent_id=self.agent_id)
        self._flows_tracker = Flows(
            creds=self.creds, agent_id=self.agent_id)
        self._pages_tracker = Pages(creds=self.creds)
        self._entity_tracker = EntityTypes(
            creds=self.creds, agent_id=self.agent_id)
        self._rg_tracker = TransitionRouteGroups(
            creds=self.creds, agent_id=self.agent_id
        )

    def _get_flows_map(self, agent_id: str = None):
        return self._flows_tracker.get_flows_map(agent_id)

    def _list_all_pages(self, flows):
        """Get a List of all pages from every flow."""
        pages = []
        for flow in flows:
            pages += self._pages_tracker.list_pages(flow)
        return pages

    def _list_all_rgs(self, flows):
        """Get a list of all route groups from every flow."""
        rgs = []
        for flow in flows:
            rgs += self._rg_tracker.list_transition_route_groups(flow)

        return rgs

    def get_agent_stats(self, agent_id: str = None, output="stdout"):
        """Provides a snapshot of resource stats from the specified CX Agent

        Pulls all resources from CX Agent and iterates over Flows/Pages to calc
        various design-time stats to be used for offline analysis, determining
        bot complexity, and various other tasks.

        Args:
          agent_id: the CX Agent ID to pull stats from
          output: Optional output format of the stats which can be ONEOF:
            'stdout', 'dict'
        """

        if not agent_id:
            agent_id = self.agent_id

        agent_obj = self._agents_tracker.get_agent(agent_id)

        flows_map = self._get_flows_map(agent_id)

        all_intents = self._intents_tracker.bulk_intent_to_df(
            agent_id=agent_id)
        all_entity_types = self._entity_tracker.list_entity_types(
            agent_id=agent_id)
        all_pages = self._list_all_pages(flows_map)
        all_rgs = self._list_all_rgs(flows_map)

        agent_display_name = agent_obj.display_name
        flows_count = len(flows_map.keys())
        pages_count = len(all_pages)
        intents_count = all_intents.display_name.nunique()
        tp_count = all_intents.shape[0]
        entity_types_count = len(all_entity_types)
        rg_count = len(all_rgs)

        if output == "stdout":
            print(f"Agent ID: {agent_id}")
            print(f"Agent Display Name: {agent_display_name}")
            print(f"Total # of Flows: {flows_count}")
            print(f"Total # of Pages: {pages_count}")
            print(f"Total # of Intents: {intents_count}")
            print(f"Total # of Training Phrases: {tp_count}")
            print(f"Total # of Entity Types: {entity_types_count}")
            print(f"Total # of Route Groups: {rg_count}")

        if output == "dict":
            stats = {
                "agent_id": agent_id,
                "display_name": agent_display_name,
                "flows": flows_count,
                "pages": pages_count,
                "intents": intents_count,
                "training_phrases": tp_count,
                "entity_types": entity_types_count,
                "route_groups": rg_count
            }

            return stats

        return None



"""Test Cases Resource Async functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List

from google.cloud.dialogflowcx_v3beta1 import services
from  google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core.scrapi_base import ScrapiBase

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TestCasesAsync(ScrapiBase):
    """Core Class for CX Test Cases Async."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id
            self.client_options = self._set_region(self.agent_id)

    @staticmethod
    def iterator(response, attribute):
        """Loops through pages of results to give all results"""
        iterated = []
        for instance in getattr(response, attribute):
            iterated.append(instance)
        return iterated

    def list_test_cases(self, agent_id: str = None):
        """Lists all Test Cases for a given Agent.

        Args:
          agent_id: The agent to list all pages for. Format:
             `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          List of test cases from an agent as a coroutine
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.ListTestCasesRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)

        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_test_cases(request)

        return response

    def export_test_cases(
        self,
        gcs_uri: str,
        agent_id: str = None,
        data_format: str = None,
        data_filter: str = None,
    ):
        """Export test cases from an agent to cloud storage.

        Args:
          gcs_uri: The `Google Cloud Storage URI to export the test cases to.
            The format of this URI must be `gs://<bucket-name>/<object-name>`.
            If unspecified, the serialized test cases is returned inline.
          agent_id: Optional. The agent where to export test cases from.
            Format: `projects/<Project ID>/locations/<Location ID>/agents/
              <Agent ID>`
          data_format: The data format of the exported test cases. If not
            specified, ``BLOB`` is assumed.
          data_filter: The filter expression used to filter exported test
            cases, see `API Filtering <https://aip.dev/160>`__. The expression
            is case insensitive and supports the following syntax:
            name = [OR name = ] ...
            For example:
              -  "name = t1 OR name = t2" matches the test case with the
              exact resource name "t1" or "t2".

        Returns:
          Long running operation for export as a coroutine
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.ExportTestCasesRequest()
        request.parent = agent_id
        request.gcs_uri = gcs_uri
        request.data_format = data_format
        request.filter = data_filter

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_test_cases(request)
        return response

    def create_test_case(
        self,
        test_case: types.TestCase,
        agent_id: str = None):
        """Create a new Test Case.

        Args:
          test_case: The Test Case to create.
          agent_id: The agent to create the test case for. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          Test case which was created as a coroutine
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.CreateTestCaseRequest()
        request.parent = agent_id
        request.test_case = test_case

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_test_case(request)
        return response

    def get_test_case(self, test_case_id: str):
        """Get the specified Test Case object.

        Args:
          test_case_id: The name of the testcase. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>`.

        Returns:
          Test Case object as coroutine
        """

        request = types.test_case.GetTestCaseRequest()
        request.name = test_case_id

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_test_case(request)
        return response

    def import_test_cases(self, gcs_uri: str, agent_id: str = None):
        """Import Test Cases from Google Cloud Storage.

        Args:
          gcs_uri: The Google Cloud Storage URI to import test cases from. The
            format of this URI must be `gs://<bucket-name>/<object-name>`.
          agent_id: The agent to import test cases to. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          Long running operation for importing test cases as a couroutine.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.ImportTestCasesRequest()
        request.parent = agent_id
        request.gcs_uri = gcs_uri

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.import_test_cases(request)
        return response

    def batch_delete_test_cases(
        self,
        test_case_ids: List[str],
        agent_id: str = None):
        """Delete a set of Test Cases from an Agent.

        Args:
          test_case_ids: List of Test Case IDs to Delete. Format of Test Case
            IDs: `projects/<Project ID>/locations/ <Location ID>/agents/
              <AgentID>/testCases/<TestCase ID>`.
          agent_id: The agent to delete test cases from. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          None
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.BatchDeleteTestCasesRequest()
        request.parent = agent_id
        request.names = test_case_ids

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        client.batch_delete_test_cases(request)

    def list_test_case_results(self, test_case_id: str):
        """List a set of Test Case results for a given Test Case ID.

        Args:
          test_case_id: The test case to list results for. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>`. Specify a ``-`` as a wildcard for
              TestCase ID to list results across multiple test cases.

        Returns:
          List of Test Case results
        """

        request = types.test_case.ListTestCaseResultsRequest()
        request.parent = test_case_id

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_test_case_results(request)
        return response

    def batch_run_test_cases(
        self,
        test_cases: List[str],
        agent_id: str = None,
        environment: str = None):
        """Run a set of Test Cases to get their latest results.

        Args:
          test_cases: List of Test Case IDs to run.
          agent_id: Agent name. Format: `projects/<Project ID>/locations/
            <Location ID>/agents/<AgentID>`.
          environment: If not set, DRAFT environment is assumed. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
            environments/<Environment ID>`.

        Returns:
          Results for the set of Test Cases as a coroutine.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.BatchRunTestCasesRequest()
        request.parent = agent_id
        request.environment = environment
        request.test_cases = test_cases
        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.batch_run_test_cases(request)
        return response

    def update_test_case(
        self,
        obj: types.TestCase,
        test_case_id: str = None,
        **kwargs) -> types.TestCase:
        """Update Test Case attributes for a specified Test Case.

        Args:
          obj: The new Test Case object to use for update.
          test_case_id: (Optional) The Test Case ID to update.

        Returns:
          Updated test case as coroutine.
        """
        test_case = obj
        if test_case_id:
            test_case.name = test_case_id

        for key, value in kwargs.items():
            setattr(test_case, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        request = types.test_case.UpdateTestCaseRequest()
        request.test_case = test_case
        request.update_mask = mask

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_test_case(request)
        return response

    def run_test_case(self, test_case_id: str, environment: str = None):
        """Run test case and get result for a specified test case.

        Args:
          test_case_id: Test Case ID to run with format:
            `projects/<Project ID>/locations/ <Location ID>/agents/<AgentID>/
              testCases/<TestCase ID>`.
          environment: (Optional) Environment name. If not set, DRAFT
            environment is assumed. Format: `projects/<Project ID>/locations/
              <Location ID>/agents/<Agent ID>/environments/<Environment ID>`.

        Returns:
          Test Case result as coroutine
        """

        request = types.test_case.RunTestCaseRequest()
        request.name = test_case_id
        request.environment = environment

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.run_test_case(request)
        return response

    def get_test_case_result(self, test_case_result_id: str):
        """Get Test Case result for a specified run on a specified Test Case.

        Args:
          test_case_result_id: The ID of the Test Case Result. Format:
            projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>/results/<TestCaseResult ID>

        Returns:
          Test Case result as coroutine.
        """

        request = types.test_case.GetTestCaseResultRequest()
        request.name = test_case_result_id

        client_options = self._set_region(test_case_result_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_test_case_result(request)
        return response

    def calculate_coverage(self, coverage_type: int, agent_id: str = None):
        """Calculate coverage of different resources in the test case set.

        Args:
          coverage_type: The type of coverage requested.
            INTENT = 1
            PAGE_TRANSITION = 2
            TRANSITION_ROUTE_GROUP = 3
          agent_id: The agent to calculate coverage for. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          Coroutine for returning the coverage of the test cases for
            the type_ specified.
        """

        if not agent_id:
            agent_id = self.agent_id

        if coverage_type not in [1, 2, 3]:
            raise ValueError(
                f"invalid coverage_type: {coverage_type}. coverage_type must "
                "be must be 1, 2 or 3"
            )

        request = types.test_case.CalculateCoverageRequest()
        request.agent = agent_id
        request.type_ = coverage_type

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesAsyncClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.calculate_coverage(request)
        return response



"""DFCX End to End Conversation Functions"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import traceback
import uuid

from typing import Dict, Any
from operator import attrgetter
from threading import Thread

import pandas as pd

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.api_core import exceptions as core_exceptions
from proto.marshal.collections import repeated
from proto.marshal.collections import maps

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import flows
from dfcx_scrapi.core import pages

logging.basicConfig(
    format="[dfcx] %(levelname)s:%(message)s", level=logging.INFO
)

MAX_RETRIES = 3


class DialogflowConversation(scrapi_base.ScrapiBase):
    """Class that wraps the SessionsClient to hold end to end conversations
    and maintain internal session state
    """

    def __init__(
        self,
        config=None,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        agent_id: str = None,
        language_code: str = "en",
    ):

        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            agent_id=agent_id,
        )

        logging.debug(
            "create conversation with creds_path: %s | agent_id: %s",
            creds_path, agent_id)

        self.agent_id = self._set_agent_id(agent_id, config)
        self.language_code = self._set_language_code(language_code, config)
        self.start_time = None
        self.query_result = None
        self.session_id = None
        self.turn_count = None
        self.agent_env = {}  # empty
        self.restart()
        self.flows = flows.Flows(creds=self.creds)
        self.pages = pages.Pages(creds=self.creds)

    @staticmethod
    def _set_language_code(language_code: str, config: Dict[str, Any]) -> str:
        """Determines how to set the language_code based on user inputs.

        We implement this for backwards compatability.
        """
        # Config will take precedence if provided
        if config:
            config_lang_code = config.get("language_code", None)

            # We'll only return if it exist in the config on the off chance that
            # some users have provided the langauge_code as a top level arg in
            # addition to providing the config
            if config_lang_code:
                return config_lang_code

        return language_code

    @staticmethod
    def _set_agent_id(input_agent_id: str, config: Dict[str, Any]) -> str:
        """Determines how to set the agent_id based on user inputs.

        We implement this for backwards compatability.
        """

        # Config will take precedence if provided
        if config:
            config_agent_path = config.get("agent_path", None)

            # We'll only return if it exist in the config on the off chance that
            # some users have provided the agent_id as a top level arg in
            # addition to providing the config
            if config_agent_path:
                return config_agent_path

        elif input_agent_id:
            return input_agent_id

        return None

    @staticmethod
    def _get_match_type_from_map(match_type: int):
        """Translates the match_type enum int value into a more descriptive
        string.
        """
        match_type_map = {
            0: "MATCH_TYPE_UNSPECIFIED",
            1: "INTENT",
            2: "DIRECT_INTENT",
            3: "PARAMETER_FILLING",
            4: "NO_MATCH",
            5: "NO_INPUT",
            6: "EVENT",
            8: "KNOWLEDGE_CONNECTOR"
        }

        return match_type_map[match_type]

    @staticmethod
    def _validate_test_set_input(test_set: pd.DataFrame):
        """Validates that all pages referenced in the test set exist in the
        agent.
        """
        mask = test_set.page_id.isna().to_list()
        invalid_pages = set(test_set.page_display_name[mask].to_list())

        if invalid_pages:
            raise UserWarning(
                "The following Pages are invalid and missing Page "
                f"IDs: \n{invalid_pages}\n\nPlease ensure that your Page "
                "Display Names do not contain typos.\nFor Default Start Page "
                "use the special page display name START_PAGE."
            )

    @staticmethod
    def progress_bar(current, total, bar_length=50, type_="Progress"):
        """Display progress bar for processing."""
        percent = float(current) * 100 / total
        arrow = "-" * int(percent / 100 * bar_length - 1) + ">"
        spaces = " " * (bar_length - len(arrow))
        print(
            f"{type_}({current}/{total})" + f"[{arrow}{spaces}] {percent:.2f}%",
            end="\r",
        )

    @staticmethod
    def _build_query_params_object(parameters, current_page, disable_webhook):
        if parameters:
            query_params = types.session.QueryParameters(
                disable_webhook=disable_webhook,
                parameters=parameters,
                current_page=current_page,
            )
        else:
            query_params = types.session.QueryParameters(
                disable_webhook=disable_webhook, current_page=current_page
            )

        return query_params

    @staticmethod
    def _build_query_input_object(input_obj, language_code):
        if "dtmf" in input_obj:
            digits = str(input_obj["dtmf"])

            finish_digit = None
            if "finish_digit" in input_obj:
                finish_digit = str(input_obj["finish_digit"])

            dtmf_input = types.session.DtmfInput(
                digits=digits, finish_digit=finish_digit
            )
            query_input = types.session.QueryInput(
                dtmf=dtmf_input,
                language_code=language_code,
            )

        elif "intent" in input_obj:
            intent_input = types.session.IntentInput(intent=input_obj["intent"])
            query_input = types.session.QueryInput(
                intent=intent_input, language_code=language_code
            )

        elif "event" in input_obj:
            event_input = types.session.EventInput(event=input_obj["event"])
            query_input = types.session.QueryInput(
                event=event_input, language_code=language_code
            )

        elif "text" in input_obj:
            text = input_obj["text"]
            logging.debug("Input text: %s", text)
            text_input = types.session.TextInput(text=text)
            query_input = types.session.QueryInput(
                text=text_input,
                language_code=language_code,
            )

        return query_input

    @staticmethod
    def _gather_text_responses(text_message):

        flat_texts = "\n".join(text_message.text)

        return flat_texts

    def _gather_response_messages(self, response_messages):
        rm_gathered = []
        for msg in response_messages:
            if msg.payload:
                msg = {
                    "payload": self.recurse_proto_marshal_to_dict(msg.payload)
                }

            elif msg.play_audio:
                msg = {"play_audio": {"audio_uri": msg.play_audio.audio_uri}}

            elif msg.live_agent_handoff:
                msg = {
                    "live_agent_handoff": self.recurse_proto_marshal_to_dict(
                        msg.live_agent_handoff.metadata
                    )
                }

            elif msg.conversation_success:
                msg = {
                    "conversation_success": self.recurse_proto_marshal_to_dict(
                        msg.conversation_success.metadata
                    )
                }

            elif msg.output_audio_text:
                msg = {"output_audio_text": msg.output_audio_text.text}

            elif msg.text:
                msg = {"text": self._gather_text_responses(msg.text)}

            rm_gathered.append(msg)

        return rm_gathered

    def _gather_query_result_parameters(self, input_parameters):
        output_parameters = {}
        for param in input_parameters:
            val = input_parameters[param]

            # If we find a RepeatedComposite (i.e. List) we will recurse
            # down and convert to lists/dics/str as needed.
            if isinstance(val, repeated.RepeatedComposite):
                val = self.recurse_proto_repeated_composite(val)

            elif isinstance(val, maps.MapComposite):
                val = self.recurse_proto_marshal_to_dict(val)

            output_parameters[param] = val

        return output_parameters

    def _page_id_mapper(self):
        """Initializes the agent_pages_map dataframe.

        This dataframe contains the flow_display_name, page_display_name,
        and page_id for each page in the agent.
        """
        agent_pages_map = pd.DataFrame()
        flow_map = self.flows.get_flows_map(agent_id=self.agent_id)
        for flow_id in flow_map.keys():

            page_map = self.pages.get_pages_map(flow_id=flow_id)

            flow_mapped = pd.DataFrame.from_dict(page_map, orient="index")
            flow_mapped["page_id"] = flow_mapped.index

            flow_mapped = flow_mapped.rename(columns={0: "page_display_name"})

            # add start page
            start_page_id = flow_id + "/pages/START_PAGE"
            flow_mapped = pd.concat(
                [
                    flow_mapped,
                    pd.DataFrame(
                        columns=["page_display_name", "page_id"],
                        data=[["START_PAGE", start_page_id]],
                    ),
                ]
            )

            flow_mapped.insert(0, "flow_display_name", flow_map[flow_id])
            agent_pages_map = pd.concat([agent_pages_map, flow_mapped])

        self.agent_pages_map = agent_pages_map.reset_index(drop=True)

    def _get_reply_results(self, utterance, page_id, results, i):
        """Get results of single text utterance to CX Agent.

        Args:
          utterance: Text to send to the bot for testing.
          page_id: Specified CX Page to send the utterance request to
          results: Pandas Dataframe to capture and store the results
          i: Internal tracking for Python Threading
        """
        response = self.reply(
            send_obj={"text": utterance}, current_page=page_id, restart=True
        )

        target_page = response["page_name"]

        results["target_page"][i] = target_page
        results["match"][i] = response["match"]

    def _get_intent_detection(self, test_set: pd.DataFrame):
        """Gets the results of a subset of Intent Detection tests.

        NOTE - This is an internal method used by run_intent_detection to
        manage parallel intent detection requests and should not be used as a
        standalone function.
        """

        self._page_id_mapper()
        test_set_mapped = pd.merge(
            test_set,
            self.agent_pages_map,
            on=["flow_display_name", "page_display_name"],
            how="left",
        )
        utterances = list(test_set_mapped["utterance"])
        page_ids = list(test_set_mapped["page_id"])

        self._validate_test_set_input(test_set_mapped)

        threads = [None] * len(utterances)
        results = {
            "target_page": [None] * len(utterances),
            "match":[None] * len(utterances),
        }
        for i, (utterance, page_id) in enumerate(zip(utterances, page_ids)):
            threads[i] = Thread(
                target=self._get_reply_results,
                args=(utterance, page_id, results, i),
            )
            threads[i].start()

        for _, thread in enumerate(threads):
            thread.join()

        test_set_mapped["target_page"] = results["target_page"]
        test_set_mapped["match"] = results["match"]
        test_set_mapped = test_set_mapped.drop(columns=["page_id"])
        intent_detection = test_set_mapped.copy()

        return intent_detection


    def restart(self):
        """Starts a new session/conversation for this agent"""
        self.session_id = uuid.uuid4()
        self.turn_count = 0

    def set_agent_env(self, param, value):
        """Setting changes related to the environment"""
        logging.info("setting agent_env param:[%s] = value:[%s]", param, value)
        self.agent_env[param] = value

    def checkpoint(self, msg=None, start=False):
        """Print a checkpoint to time progress and debug bottleneck"""
        if start:
            start_time = time.perf_counter()
            self.start_time = start_time
        else:
            start_time = self.start_time
        duration = round((time.perf_counter() - start_time), 2)
        if duration > 2:
            if msg:
                print(f"{duration:0.2f}s {msg}")

    @scrapi_base.api_call_counter_decorator
    def reply(
        self,
        send_obj: Dict[str, str],
        restart: bool = False,
        retries: int = 0,
        current_page: str = None,
        checkpoints: bool = False,
    ):
        """Runs intent detection on one utterance and gets the agent reply.

        Args:
          send_obj: Dictionary with the following structure:
            {'text': str,
            'params': Dict[str,str],
            'dtmf': str}
          restart: Boolean flag that determines whether to use the existing
            session ID or start a new conversation with a new session ID.
            Passing True will create a new session ID on subsequent calls.
            Defaults to False.
          retries: used for recurse calling this func if API fails
          current_page: Specify the page id to start the conversation from
          checkpoints: Boolean flag to enable/disable Checkpoint timer
            debugging. Defaults to False.

        Returns:
          A dictionary for the agent reply to to the submitted text.
            Includes keys response_messages, confidence, page_name,
            intent_name, match_type, match, and params.
        """
        text = send_obj.get("text")
        send_params = send_obj.get("params")

        if not text:
            logging.warning(f"Input Text is empty. {send_obj}")

        if text and len(text) > 256:
            logging.warning(
                "Text input is too long. Truncating to 256 characters."
            )
            text = text[0:256]
            logging.warning(f"TRUNCATED TEXT: {text}")

        custom_environment = self.agent_env.get("environment")
        disable_webhook = self.agent_env.get("disable_webhook") or False

        if checkpoints:
            self.checkpoint(start=True)

        if restart:
            self.restart()

        client_options = self._set_region(self.agent_id)
        session_client = services.sessions.SessionsClient(
            credentials=self.creds, client_options=client_options
        )
        session_path = f"{self.agent_id}/sessions/{self.session_id}"

        if custom_environment:
            logging.info("req using env: %s", custom_environment)
            session_path = (
                f"{self.agent_id}/environments/"
                f"{custom_environment}/sessions/{self.session_id}"
            )

        # Build Query Params object
        query_params = self._build_query_params_object(
            send_params, current_page, disable_webhook
        )

        # Build Query Input object
        query_input = self._build_query_input_object(
            send_obj, self.language_code
        )

        request = types.session.DetectIntentRequest(
            session=session_path,
            query_input=query_input,
            query_params=query_params,
        )

        logging.debug("query_params: %s", query_params)
        logging.debug("request %s", request)

        response = None
        try:
            response = session_client.detect_intent(request=request)

        except core_exceptions.InternalServerError as err:
            logging.error(
                "---- ERROR --- InternalServerError caught on CX.detect %s", err
            )
            logging.error("text: %s", text)
            logging.error("query_params: %s", query_params)
            logging.error("query_input: %s", query_input)
            return {}

        except core_exceptions.ClientError as err:
            logging.error(
                "---- ERROR ---- ClientError caught on CX.detect %s", err
            )
            template = "An exception of type {0} occurred. \nArguments:\n{1!r}"
            message = template.format(type(err).__name__, err.args)
            logging.error("err name %s", message)

            logging.error("text %s", text)
            logging.error("query_params %s", query_params)
            logging.error("query_input %s", query_input)
            logging.error(traceback.print_exc())
            retries += 1
            if retries < MAX_RETRIES:
                logging.error("retrying")
                return self.reply(send_obj, restart=restart, retries=retries)
            else:
                logging.error("MAX_RETRIES exceeded")
                return {}

        if checkpoints:
            self.checkpoint("<< got response")
        query_result = response.query_result
        logging.debug("dfcx>qr %s", query_result)
        self.query_result = query_result
        reply = {}

        # Gather Response Messages into List of Dicts
        if query_result.response_messages:
            response_messages = self._gather_response_messages(
                query_result.response_messages
            )
        else:
            response_messages = None

        # Convert params structures from Proto to standard python data types
        if query_result.parameters:
            params = self._gather_query_result_parameters(
                query_result.parameters
            )
        else:
            params = None

        reply["response_messages"] = response_messages
        reply["confidence"] = query_result.intent_detection_confidence
        reply["page_name"] = query_result.current_page.display_name
        reply["intent_name"] = query_result.intent.display_name
        reply["match_type"] = self._get_match_type_from_map(
            query_result.match.match_type
        )
        reply["match"] = query_result.match
        reply["params"] = params

        logging.debug("reply %s", reply)

        return reply

    def getpath(self, obj, xpath, default=None):
        """Get data at a pathed location out of object internals"""
        elem = obj
        try:
            for xpitem in xpath.strip("/").split("/"):
                try:
                    xpitem = int(xpitem)
                    elem = elem[xpitem]  # dict
                except ValueError:
                    elem = elem.get(xpitem)  # array
        except KeyError:
            logging.warning("failed to getpath: %s ", xpath)
            return default

        logging.info("OK getpath: %s", xpath)
        if self:
            return elem

        return None

    def run_intent_detection(
        self,
        test_set: pd.DataFrame,
        chunk_size: int = 300,
        rate_limit: float = 20,
    ):
        """Tests a set of utterances for intent detection against a CX Agent.

        This function uses Python Threading to run tests in parallel to
        expedite intent detection testing for Dialogflow CX agents. The default
        quota for Text requests/min is 1200. Ref:
          https://cloud.google.com/dialogflow/quotas#table

        Args:
          test_set: A Pandas DataFrame with the following schema.
            flow_display_name: str
            page_display_name: str
              - NOTE, when using the Default Start Page of a Flow you must
                define it as the special display name START_PAGE
            utterance: str
          chunk_size: Determines the number of text requests to send in
            parallel. This should be adjusted based on your test_set size and
            the Quota limits set for your GCP project. Default is 300.
          rate_limit: Number of seconds to wait between running test set chunks

        Returns:
          A Pandas DataFrame consisting of the original
            DataFrame plus an additional column for the detected intent with
            the following schema.
              flow_display_name: str
              page_display_name: str
              utterance: str
              detected_intent: str
              confidence: float
              target_page: str
        """

        result = pd.DataFrame()
        for start in range(0, test_set.shape[0], chunk_size):
            test_set_chunk = test_set.iloc[start : start + chunk_size]
            result_chunk = self._get_intent_detection(test_set=test_set_chunk)
            result = pd.concat([result, result_chunk])
            self.progress_bar(start, test_set.shape[0])
            time.sleep(rate_limit)
        self.progress_bar(test_set.shape[0], test_set.shape[0])

        result = self._unpack_match(result)
        return result

    def _unpack_match(self, df: pd.DataFrame):
        """ Unpacks a 'match' column into four component columns.

        Args:
          df: dataframe containing a column named match of types.Match

        Returns:
          A copy of df with columns match_type, confidence, parameters_set,
            and detected_intent instead of match.
        """
        df = (
            df
            .copy()
            .assign(
                match_type = lambda df: df.match.apply(
                    attrgetter("match_type._name_")),
                confidence = lambda df: df.match.apply(
                    attrgetter("confidence")),
                parameters_set = lambda df: df.match.apply(
                    attrgetter("parameters")),
                detected_intent = lambda df: df.match.apply(
                    attrgetter("intent.display_name"))
            )
            .assign(
                parameters_set = lambda df: df.parameters_set.apply(
                    lambda p: self.recurse_proto_marshal_to_dict(
                        p) if p else "")
            )
            .drop(columns="match")
        )
        return df



"""A collection of Methods to support the Change History feature in DFCX."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging

from typing import Dict

import pandas as pd

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Changelogs(scrapi_base.ScrapiBase):
    """Tools class that contains methods to support Change History feature."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id=None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id

    @staticmethod
    def _validate_create_time(create_time: str):
        """Validates that create_time is in the ISO 8601 datetime format."""
        try:
            datetime.datetime.strptime(create_time, "YYYY-MM-DDThh:mm:ss.sZ")
        except ValueError as err:
            print("Create Time should be of format: 'YYYY-MM-DDThh:mm:ss.sZ'")
            raise ValueError from err
        else:
            return True

    @staticmethod
    def _validate_epoch_time(create_time_epoch_seconds: str):
        """Validates that create_time_epoch_seconds is a Unix Timestamp."""
        try:
            datetime.datetime.fromtimestamp(int(create_time_epoch_seconds))
        except ValueError as err:
            print("Create Time should be valid Unix Timestamp")
            raise ValueError from err
        else:
            return True

    @scrapi_base.api_call_counter_decorator
    def list_changelogs(self, agent_id: str = None, **kwargs):
        """Lists all Change History logs for a CX Agent.

        This method supports log filtering via **kwargs input. The filters
        currently supported are: user_email, resource, display_name, type,
        action, and create_time.
        See https://github.com/googleapis/python-dialogflow-cx/blob/main/
        google/cloud/dialogflowcx_v3beta1/types/changelog.py#L40 for
        pointers on filter examples

        Args:
          agent_id: the formatted CX Agent ID

        Returns:
          List of Change History logs
        """
        request = types.changelog.ListChangelogsRequest()
        request.parent = agent_id

        if kwargs.items():
            filter_list = []
            for key, value in kwargs.items():
                if key == "user_email":
                    filter_list.append(f'user_email = "{value}"')
                elif key == "resource":
                    filter_list.append(f'resource = "{value}"')
                elif key == "display_name":
                    filter_list.append(f'display_name = "{value}"')
                elif key == "type":
                    filter_list.append(f'type = "{value}"')
                elif key == "action":
                    filter_list.append(f'action = "{value}"')
                elif key == "create_time":
                    pass
                    # BUG (pmarlow): Time filters not being accepted properly
                    # TODO (pmarlow): implement input validation
                    # filter_list.append(f"create_time {value}")
                elif key == "create_time_epoch_seconds":
                    pass
                    # BUG (pmarlow): Time filters not being accepted properly
                    # TODO (pmarlow): implement input validation
                    # filter_list.append(
                    # f"\"create_time_epoch_seconds {value}\"")

            if len(filter_list) < 1:
                pass

            elif len(filter_list) == 1:
                filter_str = filter_list[0]
                request.filter = filter_str

            else:
                filter_str = filter_list[0]
                for item in filter_list[1:]:
                    filter_str += f" AND {item}"
                request.filter = filter_str

        client_options = self._set_region(agent_id)
        client = services.changelogs.ChangelogsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_changelogs(request)

        changelogs = []
        for page in response.pages:
            for log in page.changelogs:
                changelogs.append(log)

        return changelogs

    @scrapi_base.api_call_counter_decorator
    def get_changelog(self, changelog_id: str):
        """Get a single changelog resource object.

        Args:
          changelog_id: The ID of the changelog to get. Format: `projects/
            <Project ID>/locations/<Location ID>/agents/<Agent ID>/changelogs/
            <Changelog ID>`

        Returns:
          A single changelog object
        """
        request = types.changelog.GetChangelogRequest()
        request.name = changelog_id

        client_options = self._set_region(changelog_id)
        client = services.changelogs.ChangelogsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_changelog(request)

        return response

    def changelogs_to_dataframe(
        self, agent_id: str, email_pattern: str = "@google.com"
    ):
        """Format the output of list_changelogs into a Pandas Dataframe.

        Args:
          agent_id: the formatted CX Agent ID
          email_pattern: Provides a match filter that will determine the value
            of the 'user_type' column. Defaults to '@google.com'.

        Returns:
          The final dataframe output of the formatted logs
        """
        changelogs = self.list_changelogs(agent_id)

        if not changelogs:
            return print("No Change History Results for this Agent.")

        df = pd.DataFrame()

        for log in changelogs:
            if email_pattern in log.user_email:
                user_type = "Internal"
            elif log.type_ == "backups":
                user_type = "System"
            else:
                user_type = "External"

            log_data = pd.DataFrame(
                columns=[
                    "create_time",
                    "display_name",
                    "resource_type",
                    "action",
                    "user_email",
                    "user_type",
                    "resource_id",
                    "changelog_id",
                ],
                data=[
                    [
                        log.create_time.strftime("%Y-%m-%d %H:%M:%S"),
                        log.display_name,
                        log.type_.title()[:-1],
                        log.action,
                        log.user_email,
                        user_type,
                        log.resource,
                        log.name,
                    ]
                ],
            )

            df = df.append(log_data)

        df = df.reset_index(drop=True)

        return df



"""Utility file for utterance generator to work with CX resources."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import string
from typing import List, Dict

import pandas as pd

from google.oauth2 import service_account
from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import intents
from dfcx_scrapi.core_ml import utterance_generator

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class UtteranceGeneratorUtils(scrapi_base.ScrapiBase):
    """Wrapper for utterance generator that creates new training phrases.

    Can be used to create independent test sets and net-new training phrases
    for intents.
    """

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds: service_account.Credentials = None,
        scope=False,
    ):

        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        logging.info("setting up utils....")
        self.intents = intents.Intents(creds_path, creds_dict)
        logging.info("downloading model....")
        self.utterance_generator = utterance_generator.UtteranceGenerator()
        logging.info("utterance generator utils setup")

    @staticmethod
    def _progress_bar(
        current: int, total: int, bar_length: int = 50, type_: str = "Progress"
    ):
        """Display progress bar for processing.

        Args:
          current: number for current iteration.
          total: number for total iterations.
          bar_length: number of spaces to make the progress bar, default 50.
          type_: label for the bar, default 'Progress'.
        """
        percent = float(current) * 100 / total
        arrow = "-" * int(percent / 100 * bar_length - 1) + ">"
        spaces = " " * (bar_length - len(arrow))
        print(
            f"{type_}({current}/{total})" + f"[{arrow}{spaces}] {percent}%",
            end="\r",
        )

    @staticmethod
    def _clean_string(string_raw: str) -> str:
        """Cleans a string for comparison.

        Cleans a string with the same steps for comparison whether
        the generated text exists or not, removes phrases which
        only differ by:
          case,
          punctuation, or
          leading and trailing spaces.

        Args:
          string_raw: phrase to clean

        Returns:
          Cleaned string
        """
        return (
            string_raw.translate(str.maketrans("", "", string.punctuation))
            .lower()
            .strip()
        )

    def _remove_training(
        self,
        synth_intent_dataset: pd.DataFrame,
        existing_phrases: List[str],
    ) -> pd.DataFrame:
        """Removes generated phrases that already exist as intent TPs.

        Internal function for removing generated phrases which already
        exist within intents as training phrases. This is done after applying
        clean_string to both.

        Args:
          synth_intent_dataset: dataframe containing generated training
            phrases.
          existing_phrases: list of phrases that already exist as intent
            training phrases.

        Returns:
          A dataframe of new only generated phrases.
        """
        existing_phrases_cleaned = [
            self._clean_string(phrase) for phrase in existing_phrases
        ]
        synth_intent_dataset.insert(
            0,
            "cleaned_synthetic_phrase",
            synth_intent_dataset["synthetic_phrases"].apply(
                self._clean_string
            ),
        )
        synth_intent_dataset = synth_intent_dataset.drop_duplicates(
            subset=["training_phrase", "cleaned_synthetic_phrase"]
        )

        synth_intent_dataset.insert(
            0,
            "synthetic_in_training",
            synth_intent_dataset.apply(
                lambda x: x["cleaned_synthetic_phrase"]
                in existing_phrases_cleaned,
                axis=1,
            ),
        )
        synth_intent_dataset = (
            synth_intent_dataset[
                ~(synth_intent_dataset["synthetic_in_training"])
            ]
            .drop(
                columns=["cleaned_synthetic_phrase", "synthetic_in_training"]
            )
            .reset_index(drop=True)
        )
        return synth_intent_dataset

    def _generate_phrases_intent(
        self,
        training_phrases_one_intent: pd.DataFrame,
        synthetic_phrases_per_intent: int,
    ) -> pd.DataFrame:
        """Generates new synthetic phrases.

        Main internal function for generating new synthetic phrases from
        the existing training phrases within an intent. The synthetic phrases
        are only as good as the training phrases in the intent.

        Args:
          training_phrases_one_intent: input phrases from which to generate
            new training phrases.
          synthetic_phrases_per_intent: number of phrases to generate.

        Returns:
          A DataFrame containing synthetic training phrases.
        """
        synthetic_instances = (
            synthetic_phrases_per_intent // len(training_phrases_one_intent)
        ) + 1
        existing_phrases = list(
            set(training_phrases_one_intent["training_phrase"])
        )
        if synthetic_instances == 1:
            training_phrases_one_intent = training_phrases_one_intent.sample(
                frac=1
            ).reset_index(drop=True)
            training_phrases_one_intent = training_phrases_one_intent.iloc[
                :synthetic_phrases_per_intent
            ]

        attempts = 0
        while True:
            synth_intent_dataset = self.utterance_generator.generate_utterances(
                training_phrases_one_intent,
                synthetic_instances=synthetic_instances,
            )
            # Check if exist in existing intents
            synth_intent_dataset = self._remove_training(
                synth_intent_dataset, existing_phrases
            )

            # check if dont have enough examples
            if len(synth_intent_dataset) >= (
                synthetic_phrases_per_intent - 1
            ):
                break
            synth_intent_dataset["synthetic_instances"] += 1
            attempts += 1
            if attempts > 3:
                break

        synth_intent_dataset = synth_intent_dataset.sample(
            frac=1
        ).iloc[:synthetic_phrases_per_intent]
        return synth_intent_dataset

    def _generate_phrases(
        self, training_phrases: pd.DataFrame, dataset_size: int
    ) -> pd.DataFrame:
        """Generates phrases for all user-specified intents.

        Internal function for running _generate_phrases_intent for all the
        user-specified intents.

        Args:
          training_phrases: df of training phrases for multiple intents with
            an Intent "display_name" column.
          dataset_size: number of requested phrases to generate over all
            specified intents.

        Returns:
            A DataFrame of generated training phrases.
        """
        # TODO: fix math for training phrase generation.
        synthetic_dataset = pd.DataFrame()
        intents_list = list(set(training_phrases["display_name"]))
        unique_intents_count = len(intents_list)

        synthetic_phrases_per_intent = dataset_size // unique_intents_count + 1

        i = 0
        for intent in intents_list:
            training_phrases_one_intent = training_phrases.copy()[
                training_phrases["display_name"] == intent
            ].reset_index(drop=True)
            intent_set = self._generate_phrases_intent(
                training_phrases_one_intent, synthetic_phrases_per_intent
            )
            synthetic_dataset = pd.concat([synthetic_dataset, intent_set])
            i += 1
            self._progress_bar(i, len(intents_list))

        return synthetic_dataset

    def create_synthetic_dataset(
        self, agent_id: str, intent_subset: List[str], dataset_size: int = 100
    ) -> pd.DataFrame:
        """Creates a synthetic test dataset.

        Creates a test dataset where none of the utterances in the test
        dataset are in the training of the existing phrases.

        Args:
          agent_id: ID of the DFCX agent.
          intent_subset: intents to generate a test dataset for.
          dataset_size: number of synthetic phrases to generate, default 100.

        Returns:
          DataFrame containing synthetic test dataset utterances with columns:
            id: IDs of original utterances.
            synthetic instances: number of synthetic phrases per
              original phrase.
            utterance: original utterances.
            synthetic_phrases: generated phrases.
            intent: intent the utterance is from.
        """
        training_phrases = self.intents.bulk_intent_to_df(
            agent_id=agent_id, intent_subset=intent_subset
        )
        training_phrases = training_phrases.copy().rename(
            columns={"tp": "utterance"}
        )
        test_dataset = self._generate_phrases(training_phrases, dataset_size)
        test_dataset = test_dataset[:dataset_size]

        return test_dataset.reset_index(drop=True)

    def create_test_dataset(
        self,
        agent_id: str,
        intent_subset: List[str],
        flow_display_name: str = "Default Start Flow",
        page_display_name: str = "START_PAGE",
        dataset_size: int = 100,
    ) -> pd.DataFrame:
        """Creates a test dataset for a given list of intents.

        The phrases in this dataset will not be exact string match phrases that
        exist in the training phrases but will be close semantically. This set
        is automatically labeled by the intent whose training was used to
        generate the new phrase. This can be used to run through the
        core.conversations run_intent_detection function. You may need to
        specify a flow_display_name and page_display_name in the dataframe to
        run the set at the correct location.

        Args:
          agent_id: name parameter of the agent to pull intents from
            full path to agent.
          intent_subset: display names of the intents to create a new
            phrases for, base phrases come from the training in the intent.
          flow_display_name: display name of the flow location at which
            to run the core.conversations run_intent_detection
            function, default Default Start Flow.
          page_display_name: display name of the page location at which
            to run the core.conversations run_intent_detection function,
            default START_PAGE.
          dataset_size: overall target size of the test set to create,
            may be less depending if new independent phrases can be
            generated from the data. The function tries to get even
            entries per intent, default 100.

        Returns:
          Dataframe with columns:
            flow_display_name: display name of the flow location at which
              to run the core.conversations run_intent_detection
              function.
            page_display_name: display name of the page location at which
              to run the core.conversations run_intent_detection function.
            utterance: original utterances.
        """
        synthetic_dataset = self.create_synthetic_dataset(
            agent_id, intent_subset, dataset_size
        )
        test_dataset = (
            synthetic_dataset.copy()[["synthetic_phrases", "display_name"]]
            .rename(columns={"synthetic_phrases": "utterance"})
            .reset_index(drop=True)
        )
        # Reformat test_dataset to fit run_intent_detection schema
        test_dataset["flow_display_name"] = flow_display_name
        test_dataset["page_display_name"] = page_display_name
        schema_dataset = test_dataset[
            ["flow_display_name", "page_display_name", "utterance"]
        ]

        return schema_dataset

    def create_new_training_phrases(
        self, agent_id: str, intent_subset: List[str], new_phrases: int = 100
    ) -> pd.DataFrame:
        """Creates new training phrases for a given list of intents.

        Generates phrases that are semantically similar to the input training
        phrases and returns them in a dataframe. The resulting dataframe can
        be used with the core.intents.modify_training_phrase_df method to
        create the appropriately formatted training phrase dataframe that will
        be ready to update to a CX agent. Using this newly formatted dataframe
        (and optionally a Parameters dataframe), the
        tools.dataframe_functions.bulk_update_intents_from_dataframe method can
        be used to make the final updates to the CX agent.

        Args:
          agent_id: name parameter of the agent to pull intents from - full
            path to agent.
          intent_subset: display names of the intents to create a new phrases
            for, base phrases come from the training in the intent.
          new_phrases: overall target size of new phrases to create, may be
            less depending if new independent phrases can be generated from
            the data. The function tries to get even entries per intent.

        Returns:
          Dataframe with columns:
            display_name: intent to add the phrase to.
            phrase: new phrase.
            action: "add".
        """
        synthetic_dataset = self.create_synthetic_dataset(
            agent_id, intent_subset, new_phrases
        )
        new_training = (
            synthetic_dataset.copy()[["display_name", "synthetic_phrases"]]
            .rename(columns={"synthetic_phrases": "phrase"})
            .reset_index(drop=True)
        )
        new_training.insert(len(new_training.columns), "action", "add")

        return new_training



"""Entity Types Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List

import pandas as pd
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class EntityTypes(scrapi_base.ScrapiBase):
    """Core Class for CX Entity Type Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        entity_id: str = None,
        agent_id: str = None,
        language_code: str = "en"
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.entity_id = entity_id
        self.agent_id = agent_id
        self.language_code = language_code


    @staticmethod
    def entity_type_proto_to_dataframe(
        obj: types.EntityType, mode: str = "basic"
    ):
        """Converts an EntityType protobuf object to a Pandas Dataframe.

        Args:
          obj: EntityType protobuf object
          mode: (Optional) "basic" returns display_name, value of entity type
            and its synonyms.
            "advanced" returns entity types and excluded phrases in a
            comprehensive format.

        Returns:
          In basic mode, a Pandas Dataframe for the entity type object with
          schema:
            display_name: str
            entity_value: str
            synonyms: str
          In advanced mode, a dictionary with keys entity_types and
          excluded_phrases.
            The value of entity_types is a Pandas Dataframe with columns:
              entity_type_id, display_name, kind, auto_expansion_mode,
              fuzzy_extraction, redact, entity_value, synonyms
            The value of excluded_phrases is a Dataframe with columns:
              entity_type_id, display_name, excluded_phrase
        """
        if mode == "basic":
            main_df = pd.DataFrame()

            entity_type_dict = {}
            entity_type_dict["display_name"] = obj.display_name
            for entity in obj.entities:
                entity_type_dict["entity_value"] = entity.value
                for synonym in entity.synonyms:
                    entity_type_dict["synonyms"] = synonym
                    entity_type_df = pd.DataFrame(entity_type_dict, index=[0])
                    main_df = pd.concat([main_df, entity_type_df],
                                        ignore_index=True)

            return main_df

        elif mode == "advanced":

            main_df = pd.DataFrame()
            excl_phrases_df = pd.DataFrame()

            excl_phrases_dict = {
                "entity_type_id": obj.name,
                "display_name": obj.display_name,
            }
            entity_type_dict = {
                "entity_type_id": obj.name,
                "display_name": obj.display_name,
                "kind": obj.kind.name,
                "auto_expansion_mode": obj.auto_expansion_mode,
                "fuzzy_extraction": obj.enable_fuzzy_extraction,
                "redact": obj.redact,
            }
            for entity in obj.entities:
                entity_type_dict["entity_value"] = entity.value
                for synonym in entity.synonyms:
                    entity_type_dict["synonyms"] = synonym
                    entity_type_df = pd.DataFrame(entity_type_dict, index=[0])
                    main_df = pd.concat([main_df, entity_type_df],
                                        ignore_index=True)

            for excluded_phrase in obj.excluded_phrases:
                excl_phrases_dict["excluded_phrase"] = excluded_phrase.value
                excl_phrases_dict2df = pd.DataFrame(
                                          excl_phrases_dict, index=[0])
                excl_phrases_df = pd.concat([excl_phrases_df,
                                     excl_phrases_dict2df], ignore_index=True)

            return {
                "entity_types": main_df, "excluded_phrases": excl_phrases_df
            }

        else:
            raise ValueError("Mode types: [basic, advanced]")

    def entity_types_to_df(
        self,
        agent_id: str = None,
        mode: str = "basic",
        entity_type_subset: List[str] = None) -> pd.DataFrame:
        """Extracts all Entity Types into a Pandas DataFrame.

        Args:
          agent_id, agent to pull list of entity types
          mode: (Optional) "basic" returns display_name, value of entity type
            and its synonyms.
            "advanced" returns entity types and excluded phrases in a
            comprehensive format.
          entity_type_subset: (Optional) A list of entities to pull
            If it's None, grab all the entity_types

        Returns:
          In basic mode, a Pandas Dataframe for all entity types in the agent
          with schema:
            display_name: str
            entity_value: str
            synonyms: str
          In advanced mode, a dictionary with keys entity_types and
          excluded_phrases.
            The value of entity_types is a Pandas Dataframe with columns:
              entity_type_id, display_name, kind, auto_expansion_mode,
              fuzzy_extraction, redact, entity_value, synonyms
            The value of excluded_phrases is a Dataframe with columns:
              entity_type_id, display_name, excluded_phrase
        """

        if not agent_id:
            agent_id = self.agent_id

        entity_types = self.list_entity_types(agent_id)
        if mode == "basic":
            main_df = pd.DataFrame()
            for obj in entity_types:
                if (entity_type_subset and
                        obj.display_name not in entity_type_subset):
                    continue

                single_entity_df = self.entity_type_proto_to_dataframe(
                    obj, mode=mode
                )
                main_df = pd.concat([main_df, single_entity_df])
            main_df = main_df.sort_values(
                ["display_name", "entity_value"])
            return main_df

        elif mode == "advanced":
            main_df = pd.DataFrame()
            excl_phrases_df = pd.DataFrame()
            for obj in entity_types:
                if (entity_type_subset and
                        obj.display_name not in entity_type_subset):
                    continue
                single_entity_dict = self.entity_type_proto_to_dataframe(
                    obj, mode=mode
                )
                main_df = pd.concat(
                    [main_df, single_entity_dict["entity_types"]])
                excl_phrases_df = pd.concat(
                    [excl_phrases_df, single_entity_dict["excluded_phrases"]])
            type_map = {
                "auto_expansion_mode": bool,
                "fuzzy_extraction": bool,
                "redact": bool
            }
            main_df = main_df.astype(type_map)

            return {
                "entity_types": main_df, "excluded_phrases": excl_phrases_df
            }

        else:
            raise ValueError("Mode types: [basic, advanced]")


    def get_entities_map(self, agent_id: str = None, reverse=False):
        """Exports Agent Entity Type Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Entity Type UUIDs as keys and
          Entity Type display names as values
        """
        if not agent_id:
            agent_id = self.agent_id

        if reverse:
            entities_dict = {
                entity.display_name: entity.name
                for entity in self.list_entity_types(agent_id)
            }

        else:
            entities_dict = {
                entity.name: entity.display_name
                for entity in self.list_entity_types(agent_id)
            }

        return entities_dict

    @scrapi_base.api_call_counter_decorator
    def list_entity_types(self, agent_id: str, language_code: str = "en"):
        """Returns a list of Entity Type objects.

        Args:
          agent_id: the formatted CX Agent ID to use
          language_code: Specifies the language of the Entity Types listed

        Returns:
          List of Entity Type objects
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.entity_type.ListEntityTypesRequest()
        request.parent = agent_id
        request.language_code = language_code

        client_options = self._set_region(agent_id)
        client = services.entity_types.EntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_entity_types(request)

        entities = []
        for page in response.pages:
            for entity in page.entity_types:
                entities.append(entity)

        return entities

    @scrapi_base.api_call_counter_decorator
    def get_entity_type(self, entity_id: str = None, language_code: str = "en"):
        """Returns a single Entity Type object.

        Args:
          entity_id: the formatted CX Entity ID to get
          language_code: Specifies the language of the Entity Types listed

        Returns:
          The single Entity Type object
        """
        if not entity_id:
            entity_id = self.entity_id

        client_options = self._set_region(entity_id)
        client = services.entity_types.EntityTypesClient(
            credentials=self.creds, client_options=client_options
        )
        request = types.entity_type.GetEntityTypeRequest()
        request.name = entity_id
        request.language_code = language_code

        response = client.get_entity_type(request=request)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_entity_type(
        self,
        agent_id: str = None,
        display_name: str = None,
        language_code: str = None,
        obj: types.EntityType = None,
        **kwargs
    ) -> types.EntityType:
        """Creates a single Entity Type object resource.

        Args:
          agent_id: the formatted CX Agent ID to create the object on
          display_name: Human readable display name for the EntityType
          language_code: Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language
          obj: The CX EntityType object in proper format.
            Refer to `builders.entity_types.EntityTypeBuilder` to build one.

        Returns:
          A copy of the Entity Type object created
        """
        if not agent_id:
            agent_id = self.agent_id

        # If entity_type_obj is given set entity_type to it
        if obj:
            entity_type_obj = obj
            entity_type_obj.name = ""
        else:
            if not display_name:
                raise ValueError(
                    "At least display_name or obj should be specified."
                )
            entity_type_obj = types.EntityType(
                display_name=display_name,
                kind=1
            )

            # set optional arguments to entity type attributes
            for key, value in kwargs.items():
                setattr(entity_type_obj, key, value)

        client_options = self._set_region(agent_id)
        client = services.entity_types.EntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        request = types.entity_type.CreateEntityTypeRequest()

        request.parent = agent_id
        request.entity_type = entity_type_obj

        if language_code:
            request.language_code = language_code

        response = client.create_entity_type(
            request=request
        )

        return response

    @scrapi_base.api_call_counter_decorator
    def update_entity_type(
        self,
        entity_type_id: str = None,
        obj: types.EntityType = None,
        language_code: str = None,
        **kwargs):
        """Update a single CX Entity Type object.

        Pass in a the Entity Type ID and the specified kwargs for the
        parameters in Entity Types object that you want updated. If you do not
        provide an Entity Type object, the object will be fetched based on the
        ID provided. Optionally, you can include a pre-made Entity Type object
        that will be used to replace some of the parameters in the existing
        Entity Type object as defined by the kwargs provided.

        Args:
          entity_type_id: CX Entity Type ID in proper format
          obj: (Optional) a CX Entity Type object of types.EntityType
          language_code: Language code of the intents being uploaded. Ref:
            https://cloud.google.com/dialogflow/cx/docs/reference/language

        Returns:
          A copy of the updated Entity Type object
        """

        if obj:
            entity_type = obj
            entity_type.name = entity_type_id
        else:
            if not entity_type_id:
                entity_type_id = self.entity_id
            entity_type = self.get_entity_type(entity_type_id)

        # set entity type attributes to args
        for key, value in kwargs.items():
            setattr(entity_type, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(entity_type_id)
        client = services.entity_types.EntityTypesClient(
            credentials=self.creds, client_options=client_options
        )

        request = types.entity_type.UpdateEntityTypeRequest()
        request.entity_type = entity_type
        request.update_mask = mask
        if language_code:
            request.language_code = language_code

        response = client.update_entity_type(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_entity_type(self, entity_id: str = None, obj=None) -> None:
        """Deletes a single Entity Type resouce object.

        Args:
          entity_id: the formatted CX Entity ID to delete

        Returns:
          None
        """
        if not entity_id:
            entity_id = self.entity_id

        if obj:
            entity_id = obj.name
        else:
            client_options = self._set_region(entity_id)
            client = services.entity_types.EntityTypesClient(
                credentials=self.creds, client_options=client_options
            )
            client.delete_entity_type(name=entity_id)



"""Setuptools for SCRAPI package."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='dfcx-scrapi',
    version='1.7.0',
    description='A high level scripting API for bot builders, developers, and\
      maintainers.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/GoogleCloudPlatform/dfcx-scrapi',
    author='Patrick Marlow',
    author_email='pmarlow@google.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Communications :: Chat',
        'Topic :: Software Development',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='dialogflow, cx, google, bot, chatbot, intent, dfcx, entity',
    package_dir={'':'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6, <4',
    install_requires=['google-cloud-dialogflow-cx']
)



"""Page Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1.services import pages
from google.cloud.dialogflowcx_v3beta1.types import page as gcdc_page
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Pages(scrapi_base.ScrapiBase):
    """Core Class for CX Page Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        scope=None,
        creds=None,
        page_id: str = None,
        flow_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.page_id = page_id
        self.flow_id = flow_id

    @staticmethod
    def _add_generic_pages_to_map(flow_id, pages_map, reverse):
        """Add the generic page names to each Page map.

        Dialogflow CX contains a few `special` pages names that are reserved
        and do not have UUID4 format IDs. This will take the existing page
        map and insert them in for downstream lookups.

        Args:
          flow_id: The Flow ID that contains the Pages in the pages_map
          pages_map: The existing pages_map Dict from `get_pages_map` that
            we will add the new special pages to.
          reverse: Boolean flag to swap key:value -> value:key
        """
        page_names = ["START_PAGE", "END_FLOW", "END_SESSION"]

        if reverse:
            for page in page_names:
                pages_map[page] = f"{flow_id}/pages/{page}"
        else:
            for page in page_names:
                pages_map[f"{flow_id}/pages/{page}"] = page

        return pages_map

    def get_pages_map(
        self, flow_id: str = None, reverse=False
    ) -> Dict[str, str]:
        """Exports Agent Page UUIDs and Names into a user friendly dict.

        Args:
          flow_id: the formatted CX Agent Flow ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Page UUIDs as keys and display names as values.
          If Optional reverse=True, the output will return page_name:ID mapping
          instead of ID:page_name
        """
        if not flow_id:
            flow_id = self.flow_id

        if reverse:
            pages_dict = {
                page.display_name: page.name
                for page in self.list_pages(flow_id)
            }

        else:
            pages_dict = {
                page.name: page.display_name
                for page in self.list_pages(flow_id)
            }

        pages_dict = self._add_generic_pages_to_map(
            flow_id, pages_dict, reverse
        )

        return pages_dict

    @scrapi_base.api_call_counter_decorator
    def list_pages(
        self,
        flow_id: str = None,
        language_code: str = "en") -> List[gcdc_page.Page]:
        """Get a List of all pages for the specified Flow ID.

        Args:
          flow_id: the properly formatted Flow ID string
          language_code: Specifies the language of the Pages listed. While the
            majority of contents of a Page is language agnostic, the contents
            in the "Agent Says" and similar parts of a Page are affected by
            language code.

        Returns:
          A List of CX Page objects for the specific Flow ID
        """
        request = gcdc_page.ListPagesRequest()
        request.parent = flow_id
        request.language_code = language_code

        client_options = self._set_region(flow_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_pages(request)

        cx_pages = []
        for page in response.pages:
            for cx_page in page.pages:
                cx_pages.append(cx_page)

        return cx_pages

    @scrapi_base.api_call_counter_decorator
    def get_page(self, page_id: str = None) -> gcdc_page.Page:
        """Get a single CX Page object based on the provided Page ID.

        Args:
          page_id: a properly formatted CX Page ID

        Returns:
          A single CX Page Object
        """
        if not page_id:
            page_id = self.page_id

        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_page(name=page_id)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_page(
        self, flow_id: str = None, obj: gcdc_page.Page = None, **kwargs
    ) -> gcdc_page.Page:
        """Create a single CX Page object in the specified Flow ID.

        Args:
          flow_id: the CX Flow ID where the Page object will be created
          obj: (Optional) a CX Page object of gcdc_page.Page

        Returns:
          A copy of the successful Page object that was created
        """
        if not flow_id:
            flow_id = self.flow_id

        if obj:
            page = obj
            page.name = ""
        else:
            page = gcdc_page.Page()

        for key, value in kwargs.items():
            setattr(page, key, value)

        client_options = self._set_region(flow_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.create_page(parent=flow_id, page=page)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_page(
        self, page_id: str = None, obj: gcdc_page.Page = None, **kwargs
    ) -> gcdc_page.Page:
        """Update a single CX Page object.

        Args:
          page_id: the CX Page ID to update
          obj: (Optional) a CX Page object of gcdc_page.Page

        Returns:
          A copy of the successful Page object that was created
        """
        if obj:
            page = obj
            page.name = page_id
        else:
            if not page_id:
                page_id = self.page_id
            page = self.get_page(page_id)

        for key, value in kwargs.items():
            setattr(page, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.update_page(page=page, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_page(self, page_id: str = None, force: bool = False) -> str:
        """Deletes the specified Page.

        Args:
          page_id: CX Page ID string in the following Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>/pages/<Page ID>``
          force: (Optional) This field has no effect for pages with no incoming
            transitions. If set to True, Dialogflow will remove the page,
            as well as any transitions to the page.

        Returns:
          String "Page `{page_id}` successfully deleted."
        """
        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )
        req = gcdc_page.DeletePageRequest(name=page_id, force=force)
        client.delete_page(request=req)

        return f"Page `{page_id}` successfully deleted."



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Union

from google.cloud.dialogflowcx_v3beta1.types import TransitionRoute
from google.cloud.dialogflowcx_v3beta1.types import TransitionRouteGroup
from dfcx_scrapi.builders.builders_common import BuildersCommon
from dfcx_scrapi.builders.routes import TransitionRouteBuilder

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TransitionRouteGroupBuilder(BuildersCommon):
    """Base Class for CX TransitionRouteGroup builder."""

    _proto_type = TransitionRouteGroup
    _proto_type_str = "TransitionRouteGroup"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        transition_routes_str = "\n".join([
            f"\n\n - Transition Route{i+1}:\n{str(TransitionRouteBuilder(tr))}"
            for i, tr in enumerate(self.proto_obj.transition_routes)
        ])

        return (
            f"display_name: {self.proto_obj.display_name}"
            f"\nTransitionRoutes:\n{'-'*20}\n{transition_routes_str}"
        )


    def show_transition_route_group(self):
        """Show the proto_obj information."""
        self._check_proto_obj_attr_exist()

        print(self)


    def create_new_proto_obj(
        self,
        display_name: str,
        transition_routes: Union[TransitionRoute, List[TransitionRoute]] = None,
        overwrite: bool = False
    ) -> TransitionRouteGroup:
        """Create a new TransitionRouteGroup.

        Args:
          display_name (str):
            Required. The human-readable name of the
            transition route group, unique within the flow.
            The display name can be no longer than 30 characters.
          transition_routes (TransitionRoute | List[TransitionRoute]):
            Transition routes associated with this TransitionRouteGroup.
            Refer to `builders.routes.TransitionRouteBuilder` to build one.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains a TransitionRouteGroup.

        Returns:
          A TransitionRouteGroup object stored in proto_obj.
        """
        # Types error checking
        if not (display_name and isinstance(display_name, str)):
            raise ValueError("display_name should be a nonempty string.")
        if transition_routes and not (
            isinstance(transition_routes, TransitionRoute) or
            (isinstance(transition_routes, list) and all(
                isinstance(tr, TransitionRoute) for tr in transition_routes))
        ):
            raise ValueError(
                "transition_routes should be either a TransitionRoute or"
                " a list of TransitionRoutes."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains a TransitionRouteGroup."
                " If you wish to overwrite it, pass overwrite as True."
            )
        # Create the TransitionRouteGroup
        if overwrite or not self.proto_obj:
            if not transition_routes:
                transition_routes = []
            if not isinstance(transition_routes, list):
                transition_routes = [transition_routes]
            self.proto_obj = TransitionRouteGroup(
                display_name=display_name,
                transition_routes=transition_routes
            )

        return self.proto_obj


    def add_transition_route(
        self,
        transition_routes: Union[TransitionRoute, List[TransitionRoute]]
    ) -> TransitionRouteGroup:
        """Add single or multiple TransitionRoutes to the TransitionRouteGroup.

        Args:
          transition_routes (TransitionRoute | List[TransitionRoute]):
            A single or list of TransitionRoutes to add
            to the TransitionRouteGroup existed in proto_obj.
        Returns:
          A TransitionRouteGroup object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        self._is_type_or_list_of_types(
            transition_routes, TransitionRoute, "transition_routes"
        )

        if not isinstance(transition_routes, list):
            transition_routes = [transition_routes]
        self.proto_obj.transition_routes.extend(transition_routes)

        return self.proto_obj


    def remove_transition_route(
        self,
        transition_route: TransitionRoute = None,
        intent: str = None,
        condition: str = None
    ) -> TransitionRouteGroup:
        """Remove a transition route from the TransitionRouteGroup.

        At least one of the `transition_route`, `intent`, or `condition` should
        be specfied.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute to remove from the TransitionRouteGroup.
            intent (str):
              TransitionRoute's intent that should be removed from
              the TransitionRouteGroup.
            condition (str):
              TransitionRoute's condition that should be removed from
              the TransitionRouteGroup.

        Returns:
          A TransitionRouteGroup object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        new_routes = []
        for tr in self.proto_obj.transition_routes:
            if self._match_transition_route(
                transition_route=tr, target_route=transition_route,
                intent=intent, condition=condition
            ):
                continue
            new_routes.append(tr)
        self.proto_obj.transition_routes = new_routes

        return self.proto_obj
